import asyncio
import sys
import unittest.mock as mock

import pytest
from sqlalchemy import create_engine

import sky
from sky import backends
from sky import global_user_state
from sky.utils import annotations


@pytest.fixture
def _mock_db_conn(tmp_path, monkeypatch):
    # Create a temporary database file
    db_path = tmp_path / 'state_testing.db'

    sqlalchemy_engine = create_engine(f'sqlite:///{db_path}')

    monkeypatch.setattr(global_user_state, '_SQLALCHEMY_ENGINE',
                        sqlalchemy_engine)

    global_user_state.create_table(sqlalchemy_engine)


@pytest.mark.skipif(sys.platform != 'linux', reason='Only test in CI.')
def test_enabled_clouds_empty():
    # In test environment, no cloud should be enabled.
    assert sky.global_user_state.get_cached_enabled_clouds(
        sky.clouds.cloud.CloudCapability.COMPUTE, workspace='default') == []


@pytest.mark.asyncio
async def test_cluster_event_retention_daemon():
    """Test the cluster event retention daemon runs correctly."""
    with mock.patch('sky.global_user_state.skypilot_config') as mock_config:
        with mock.patch(
                'sky.global_user_state.cleanup_cluster_events_with_retention'
        ) as mock_clean:
            with mock.patch('asyncio.sleep') as mock_sleep:
                # Configure negative retention (disabled)
                mock_config.get_nested.return_value = -1

                mock_sleep.side_effect = [None, asyncio.CancelledError()]

                # Run the daemon
                with pytest.raises(asyncio.CancelledError):
                    await sky.global_user_state.cluster_event_retention_daemon()

                # Verify cleanup was NOT called due to negative retention
                mock_clean.assert_not_called()

                # Verify sleep was called with max(-1, 3600) = 3600
                assert mock_sleep.call_count == 2
                mock_sleep.assert_any_call(3600)

                # Now test when we run retention with a positive value.
                mock_config.get_nested.return_value = 2  # every 2 hours.
                mock_sleep.side_effect = [None, None, asyncio.CancelledError()]

                with pytest.raises(asyncio.CancelledError):
                    await sky.global_user_state.cluster_event_retention_daemon()

                # Verify cleanup was called with the correct retention.
                mock_clean.assert_any_call(
                    2, sky.global_user_state.ClusterEventType.STATUS_CHANGE)
                mock_clean.assert_any_call(
                    2, sky.global_user_state.ClusterEventType.DEBUG)

                # Verify sleep was called with max(2 * 3600, 3600) = 7200
                assert mock_sleep.call_count == 5
                mock_sleep.assert_any_call(7200)


def test_add_or_update_cluster_update_only(_mock_db_conn):
    # if no record exists and existing_cluster_hash is specified
    # should raise ValueError
    with pytest.raises(ValueError):
        global_user_state.add_or_update_cluster(
            'test-cluster',
            'test-cluster',
            sky.Resources(infra='aws/us-east-1/us-east-1a',
                          instance_type='p4d.24xlarge'),
            ready=True,
            existing_cluster_hash='01230123')

    # if existing_cluster_hash is not specified, should insert a new record
    global_user_state.add_or_update_cluster(
        'test-cluster',
        'test-cluster',
        sky.Resources(infra='aws/us-east-1/us-east-1a',
                      instance_type='p4d.24xlarge'),
        ready=True)

    with pytest.raises(ValueError):
        # incorrect cluster hash filter
        global_user_state.add_or_update_cluster(
            'test-cluster',
            'test-cluster',
            sky.Resources(infra='aws/us-east-1/us-east-1a',
                          instance_type='p4d.24xlarge'),
            ready=True,
            existing_cluster_hash='01230123')

    # correct cluster hash filter
    record = global_user_state.get_cluster_from_name('test-cluster')
    global_user_state.add_or_update_cluster(
        'test-cluster',
        'test-cluster',
        sky.Resources(infra='aws/us-east-1/us-east-1a',
                      instance_type='p4d.24xlarge'),
        ready=True,
        existing_cluster_hash=record['cluster_hash'])


def test_get_clusters_cache_size(_mock_db_conn):
    """Test that calling get_clusters does not cause the LRU cache
    to grow without bound."""
    # Clear any existing request-level cache before starting
    annotations.clear_request_level_cache()

    # Create multiple real CloudVmRayResourceHandle instances with synthetic data
    handles = []
    for i in range(15):  # Create more than maxsize=10 to test cache eviction
        cluster_name = f'test-cluster-{i}'

        # Create a real handle instance
        handle = backends.CloudVmRayResourceHandle(
            cluster_name=cluster_name,
            cluster_name_on_cloud=cluster_name,
            cluster_yaml=None,  # Can be None for testing
            launched_nodes=1,
            launched_resources=sky.Resources(infra='aws/us-east-1/us-east-1a',
                                             instance_type='m5.large'),
            stable_internal_external_ips=[('10.0.0.1', '1.2.3.4')],
            stable_ssh_ports=[22],
        )

        # Add to global state
        global_user_state.add_or_update_cluster(
            cluster_name,
            handle,
            requested_resources={handle.launched_resources},
            ready=True)
        handles.append(handle)

    # Get all clusters to retrieve the handles from the database
    cluster_records = global_user_state.get_clusters()
    assert len(cluster_records) == 15, "Should have 15 clusters in database"

    # Call get_command_runners on each handle to populate the cache
    cache_call_results = []

    with mock.patch('sky.backends.backend_utils.ssh_credential_from_yaml'
                   ) as mock_ssh_creds:
        # Mock SSH credentials to avoid file system dependencies
        mock_ssh_creds.return_value = {
            'ssh_user': 'ubuntu',
            'ssh_private_key': '/tmp/fake_key'
        }

        with mock.patch(
                'sky.utils.command_runner.SSHCommandRunner.make_runner_list'
        ) as mock_make_runners:
            # Mock the command runner creation to avoid actual SSH connections
            mock_make_runners.return_value = [mock.MagicMock()]

            for i, record in enumerate(cluster_records):
                handle = record['handle']

                try:
                    # Call get_command_runners which should be cached
                    runners = handle.get_command_runners()
                    cache_call_results.append(
                        ('success', len(runners) if runners else 0))

                    # Verify we get command runners back
                    assert isinstance(
                        runners,
                        list), "Should return a list of command runners"

                except Exception as e:
                    # Track any failures for debugging
                    cache_call_results.append(('error', str(e)))

                # Check if the handle's get_command_runners method has cache_info
                if hasattr(handle.get_command_runners, 'cache_info'):
                    cache_info = handle.get_command_runners.cache_info()
                    print(f"Handle {i}: Cache size={cache_info.currsize}, "
                          f"hits={cache_info.hits}, misses={cache_info.misses}")

                    # Cache size should not exceed maxsize=10
                    assert cache_info.currsize <= 10, (
                        f"Cache size {cache_info.currsize} exceeds maxsize=10")

    # Verify that calling get_command_runners multiple times uses cache
    successful_calls = [
        result for result in cache_call_results if result[0] == 'success'
    ]
    assert len(successful_calls
              ) > 0, "At least some get_command_runners calls should succeed"

    # Test calling the same handle multiple times to verify cache hits
    if cluster_records:
        first_handle = cluster_records[0]['handle']
        with mock.patch('sky.backends.backend_utils.ssh_credential_from_yaml'
                       ) as mock_ssh_creds:
            mock_ssh_creds.return_value = {
                'ssh_user': 'ubuntu',
                'ssh_private_key': '/tmp/fake_key'
            }
            with mock.patch(
                    'sky.utils.command_runner.SSHCommandRunner.make_runner_list'
            ) as mock_make_runners:
                mock_make_runners.return_value = [mock.MagicMock()]

                # Call multiple times - should hit cache
                try:
                    first_handle.get_command_runners()
                    first_handle.get_command_runners()

                    # Verify cache behavior if cache_info is available
                    if hasattr(first_handle.get_command_runners, 'cache_info'):
                        cache_info = first_handle.get_command_runners.cache_info(
                        )
                        assert cache_info.hits > 0, "Should have cache hits for repeated calls"
                        print(
                            f"After repeated calls: hits={cache_info.hits}, misses={cache_info.misses}"
                        )
                except Exception as e:
                    print(f"Cache hit test failed: {e}")

    # Test cache clearing functionality
    initial_functions = len(annotations._FUNCTIONS_NEED_RELOAD_CACHE)
    annotations.clear_request_level_cache()

    # The number of functions tracked should remain the same, but caches should be cleared
    final_functions = len(annotations._FUNCTIONS_NEED_RELOAD_CACHE)
    assert final_functions == initial_functions, "Function count should remain the same after cache clear"

    print(
        f"Test completed successfully. Tested cache behavior with {len(cluster_records)} clusters"
    )
    print(f"Successful get_command_runners calls: {len(successful_calls)}")
    print(f"Request-level cache functions tracked: {final_functions}")


@pytest.mark.parametrize('should_mock,return_user', [
    (True, True),
    (False, True),
    (True, False),
    (False, False),
])
def test_add_or_update_user_new_user(_mock_db_conn, should_mock, return_user):
    """Test add new user."""
    user_id = f'test-new-user-{str(should_mock).lower()}-{str(return_user).lower()}'
    password = 'password123'

    def test():
        user = sky.models.User(id=user_id, name='Test User', password=password)
        result = global_user_state.add_or_update_user(user,
                                                      return_user=return_user)

        if return_user:
            was_inserted, returned_user = result
            assert was_inserted is True
            assert returned_user is not None
            assert returned_user.id == user_id
            assert returned_user.name == 'Test User'
            assert returned_user.password == password
            assert returned_user.created_at is not None
        else:
            assert result is True

        fetched_user = global_user_state.get_user(user_id)
        assert fetched_user is not None
        assert fetched_user.name == 'Test User'
        assert fetched_user.password == password

    if should_mock:
        # Mock _sqlite_supports_returning so we don't have to install an older version of SQLite or SQLAlchemy.
        with mock.patch('sky.global_user_state._sqlite_supports_returning',
                        return_value=False):
            test()
    else:
        test()


@pytest.mark.parametrize('should_mock', [True, False])
def test_add_or_update_user_update_user(_mock_db_conn, should_mock):
    """Test update user."""

    def test():
        user_id = 'test-update-user'

        # Insert
        user = sky.models.User(id=user_id, name='Original', password='old')
        was_inserted, original_user = global_user_state.add_or_update_user(
            user, return_user=True)
        assert was_inserted is True
        original_created_at = original_user.created_at

        # Update
        updated_user_data = sky.models.User(id=user_id,
                                            name='Updated',
                                            password='new')
        was_inserted, updated_user = global_user_state.add_or_update_user(
            updated_user_data, return_user=True)

        assert was_inserted is False
        assert updated_user.name == 'Updated'
        assert updated_user.password == 'new'
        assert updated_user.created_at == original_created_at

        # Verify in database
        fetched_user = global_user_state.get_user(user_id)
        assert fetched_user.name == 'Updated'
        assert fetched_user.password == 'new'

    if should_mock:
        with mock.patch('sky.global_user_state._sqlite_supports_returning',
                        return_value=False):
            test()
    else:
        test()


@pytest.mark.parametrize('should_mock', [True, False])
def test_add_or_update_user_update_partial(_mock_db_conn, should_mock):
    """Test update user without changing password."""

    def test():
        user_id = 'test-update-partial-user'

        # Insert with password
        user = sky.models.User(id=user_id, name='Original', password='old')
        global_user_state.add_or_update_user(user, return_user=False)

        # Update only name
        updated_user_data = sky.models.User(id=user_id, name='Updated')
        was_inserted, updated_user = global_user_state.add_or_update_user(
            updated_user_data, return_user=True)

        assert was_inserted is False
        assert updated_user.name == 'Updated'
        assert updated_user.password == 'old'

        fetched_user = global_user_state.get_user(user_id)
        assert fetched_user.name == 'Updated'
        assert fetched_user.password == 'old'

    if should_mock:
        with mock.patch('sky.global_user_state._sqlite_supports_returning',
                        return_value=False):
            test()
    else:
        test()


@pytest.mark.parametrize('should_mock', [True, False])
def test_add_or_update_user_name_none(_mock_db_conn, should_mock):
    """Test that user with name=None returns False without inserting."""

    def test():
        user = sky.models.User(id='test-user-none', name=None)
        result = global_user_state.add_or_update_user(user, return_user=False)
        assert result is False

    if should_mock:
        with mock.patch('sky.global_user_state._sqlite_supports_returning',
                        return_value=False):
            test()
    else:
        test()



"""Initial schema for sky config database

Revision ID: 001
Revises:
Create Date: 2025-10-21

"""
# pylint: disable=invalid-name
from alembic import op

from sky.skypilot_config import Base
from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision = '001'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    """Create initial schema for config_yaml table"""
    with op.get_context().autocommit_block():
        # Create all tables with their current schema
        db_utils.add_all_tables_to_db_sqlalchemy(Base.metadata, op.get_bind())


def downgrade():
    """Drop all tables"""
    Base.metadata.drop_all(bind=op.get_bind())



"""RunPod library wrapper for SkyPilot."""

import base64
import time
from typing import Any, Dict, List, Optional, Tuple

from sky import sky_logging
from sky.adaptors import runpod
from sky.provision import docker_utils
from sky.provision.runpod.api import commands as runpod_commands
from sky.skylet import constants
from sky.utils import common_utils

logger = sky_logging.init_logger(__name__)

GPU_NAME_MAP = {
    # AMD
    'MI300X': 'AMD Instinct MI300X OAM',

    # NVIDIA A-series
    'A100-80GB': 'NVIDIA A100 80GB PCIe',
    'A100-80GB-SXM': 'NVIDIA A100-SXM4-80GB',
    'A30': 'NVIDIA A30',
    'A40': 'NVIDIA A40',

    # NVIDIA B-series
    'B200': 'NVIDIA B200',

    # GeForce
    'RTX3070': 'NVIDIA GeForce RTX 3070',
    'RTX3080': 'NVIDIA GeForce RTX 3080',
    'RTX3080Ti': 'NVIDIA GeForce RTX 3080 Ti',
    'RTX3090': 'NVIDIA GeForce RTX 3090',
    'RTX3090Ti': 'NVIDIA GeForce RTX 3090 Ti',
    'RTX4070Ti': 'NVIDIA GeForce RTX 4070 Ti',
    'RTX4080': 'NVIDIA GeForce RTX 4080',
    'RTX4080SUPER': 'NVIDIA GeForce RTX 4080 SUPER',
    'RTX4090': 'NVIDIA GeForce RTX 4090',
    'RTX5080': 'NVIDIA GeForce RTX 5080',
    'RTX5090': 'NVIDIA GeForce RTX 5090',

    # NVIDIA H100/H200
    # Following instance is displayed as SXM at the console
    # but the ID from the API appears as HBM
    'H100-SXM': 'NVIDIA H100 80GB HBM3',
    'H100-NVL': 'NVIDIA H100 NVL',
    'H100': 'NVIDIA H100 PCIe',
    'H200-SXM': 'NVIDIA H200',

    # NVIDIA L-series
    'L4': 'NVIDIA L4',
    'L40': 'NVIDIA L40',
    'L40S': 'NVIDIA L40S',

    # Ada generation (GeForce & RTX A)
    'RTX2000-Ada': 'NVIDIA RTX 2000 Ada Generation',
    'RTX4000-Ada': 'NVIDIA RTX 4000 Ada Generation',
    'RTX4000-Ada-SFF': 'NVIDIA RTX 4000 SFF Ada Generation',
    'RTX5000-Ada': 'NVIDIA RTX 5000 Ada Generation',
    'RTX6000-Ada': 'NVIDIA RTX 6000 Ada Generation',

    # NVIDIA RTX A-series
    'RTXA2000': 'NVIDIA RTX A2000',
    'RTXA4000': 'NVIDIA RTX A4000',
    'RTXA4500': 'NVIDIA RTX A4500',
    'RTXA5000': 'NVIDIA RTX A5000',
    'RTXA6000': 'NVIDIA RTX A6000',

    # Tesla V100 variants
    'V100-16GB-FHHL': 'Tesla V100-FHHL-16GB',
    'V100-16GB-SXM2': 'Tesla V100-SXM2-16GB',
    'V100-32GB-SXM2': 'Tesla V100-SXM2-32GB',
    'V100-16GB-PCIe': 'Tesla V100-PCIE-16GB',
}


def _construct_docker_login_template_name(cluster_name: str) -> str:
    """Constructs the registry auth template name."""
    return f'{cluster_name}-docker-login-template'


def retry(func):
    """Decorator to retry a function."""

    def wrapper(*args, **kwargs):
        """Wrapper for retrying a function."""
        cnt = 0
        while True:
            try:
                return func(*args, **kwargs)
            except runpod.runpod.error.QueryError as e:
                if cnt >= 3:
                    raise
                logger.warning('Retrying for exception: '
                               f'{common_utils.format_exception(e)}.')
                time.sleep(1)

    return wrapper


# Adapted from runpod.api.queries.pods.py::QUERY_POD.
# Adding containerRegistryAuthId to the query.
_QUERY_POD = """
query myPods {
    myself {
        pods {
            id
            containerDiskInGb
            containerRegistryAuthId
            costPerHr
            desiredStatus
            dockerArgs
            dockerId
            env
            gpuCount
            imageName
            lastStatusChange
            machineId
            memoryInGb
            name
            podType
            port
            ports
            uptimeSeconds
            vcpuCount
            volumeInGb
            volumeMountPath
            runtime {
                ports{
                    ip
                    isIpPublic
                    privatePort
                    publicPort
                    type
                }
            }
            machine {
                gpuDisplayName
            }
        }
    }
}
"""


def _sky_get_pods() -> dict:
    """List all pods with extra registry auth information.

    Adapted from runpod.get_pods() to include containerRegistryAuthId.
    """
    raw_return = runpod.runpod.api.graphql.run_graphql_query(_QUERY_POD)
    cleaned_return = raw_return['data']['myself']['pods']
    return cleaned_return


_QUERY_POD_TEMPLATE_WITH_REGISTRY_AUTH = """
query myself {
    myself {
        podTemplates {
            name
            containerRegistryAuthId
        }
    }
}
"""


def _list_pod_templates_with_container_registry() -> dict:
    """List all pod templates."""
    raw_return = runpod.runpod.api.graphql.run_graphql_query(
        _QUERY_POD_TEMPLATE_WITH_REGISTRY_AUTH)
    return raw_return['data']['myself']['podTemplates']


def list_instances() -> Dict[str, Dict[str, Any]]:
    """Lists instances associated with API key."""
    instances = _sky_get_pods()

    instance_dict: Dict[str, Dict[str, Any]] = {}
    for instance in instances:
        info = {}

        info['status'] = instance['desiredStatus']
        info['name'] = instance['name']
        info['port2endpoint'] = {}

        # Sometimes when the cluster is in the process of being created,
        # the `port` field in the runtime is None and we need to check for it.
        if (instance['desiredStatus'] == 'RUNNING' and
                instance.get('runtime') and
                instance.get('runtime').get('ports')):
            for port in instance['runtime']['ports']:
                if port['isIpPublic']:
                    if port['privatePort'] == 22:
                        info['external_ip'] = port['ip']
                        info['ssh_port'] = port['publicPort']
                    info['port2endpoint'][port['privatePort']] = {
                        'host': port['ip'],
                        'port': port['publicPort']
                    }
                else:
                    info['internal_ip'] = port['ip']

        instance_dict[instance['id']] = info

    return instance_dict


def delete_pod_template(template_name: str) -> None:
    """Deletes a pod template."""
    try:
        runpod.runpod.api.graphql.run_graphql_query(
            f'mutation {{deleteTemplate(templateName: "{template_name}")}}')
    except runpod.runpod.error.QueryError as e:
        logger.warning(f'Failed to delete template {template_name}: {e} '
                       'Please delete it manually.')


def delete_register_auth(registry_auth_id: str) -> None:
    """Deletes a registry auth."""
    try:
        runpod.runpod.delete_container_registry_auth(registry_auth_id)
    except runpod.runpod.error.QueryError as e:
        logger.warning(
            f'Failed to delete registry auth {registry_auth_id}: {e} '
            'Please delete it manually.')


def _create_template_for_docker_login(
    cluster_name: str,
    image_name: str,
    docker_login_config: Optional[Dict[str, str]],
) -> Tuple[str, Optional[str]]:
    """Creates a template for the given image with the docker login config.

    Returns:
        formatted_image_name: The formatted image name.
        template_id: The template ID. None for no docker login config.
    """
    if docker_login_config is None:
        return image_name, None
    login_config = docker_utils.DockerLoginConfig(**docker_login_config)
    container_registry_auth_name = f'{cluster_name}-registry-auth'
    container_template_name = _construct_docker_login_template_name(
        cluster_name)
    # The `name` argument is only for display purpose and the registry server
    # will be splitted from the docker image name (Tested with AWS ECR).
    # Here we only need the username and password to create the registry auth.
    # TODO(tian): Now we create a template and a registry auth for each cluster.
    # Consider create one for each server and reuse them. Challenges including
    # calculate the reference count and delete them when no longer needed.
    create_auth_resp = runpod.runpod.create_container_registry_auth(
        name=container_registry_auth_name,
        username=login_config.username,
        password=login_config.password,
    )
    registry_auth_id = create_auth_resp['id']
    create_template_resp = runpod.runpod.create_template(
        name=container_template_name,
        image_name=None,
        registry_auth_id=registry_auth_id,
    )
    return login_config.format_image(image_name), create_template_resp['id']


def launch(
    cluster_name: str,
    node_type: str,
    instance_type: str,
    region: str,
    zone: str,
    disk_size: int,
    image_name: str,
    ports: Optional[List[int]],
    public_key: str,
    preemptible: Optional[bool],
    bid_per_gpu: float,
    docker_login_config: Optional[Dict[str, str]],
    *,
    network_volume_id: Optional[str] = None,
    volume_mount_path: Optional[str] = None,
) -> str:
    """Launches an instance with the given parameters.

    For CPU instances, we directly use the instance_type for launching the
    instance.

    For GPU instances, we convert the instance_type to the RunPod GPU name,
    and finds the specs for the GPU, before launching the instance.

    Returns:
        instance_id: The instance ID.
    """
    name = f'{cluster_name}-{node_type}'

    # TODO(zhwu): keep this align with setups in
    # `provision.kuberunetes.instance.py`
    setup_cmd = (
        'prefix_cmd() '
        '{ if [ $(id -u) -ne 0 ]; then echo "sudo"; else echo ""; fi; }; '
        '$(prefix_cmd) apt update;'
        'export DEBIAN_FRONTEND=noninteractive;'
        '$(prefix_cmd) apt install openssh-server rsync curl patch -y;'
        '$(prefix_cmd) mkdir -p /var/run/sshd; '
        '$(prefix_cmd) '
        'sed -i "s/PermitRootLogin prohibit-password/PermitRootLogin yes/" '
        '/etc/ssh/sshd_config; '
        '$(prefix_cmd) sed '
        '"s@session\\s*required\\s*pam_loginuid.so@session optional '
        'pam_loginuid.so@g" -i /etc/pam.d/sshd; '
        'cd /etc/ssh/ && $(prefix_cmd) ssh-keygen -A; '
        '$(prefix_cmd) mkdir -p ~/.ssh; '
        '$(prefix_cmd) chown -R $(whoami) ~/.ssh;'
        '$(prefix_cmd) chmod 700 ~/.ssh; '
        f'$(prefix_cmd) echo "{public_key}" >> ~/.ssh/authorized_keys; '
        '$(prefix_cmd) chmod 644 ~/.ssh/authorized_keys; '
        '$(prefix_cmd) service ssh restart; '
        '$(prefix_cmd) export -p > ~/container_env_var.sh && '
        '$(prefix_cmd) '
        'mv ~/container_env_var.sh /etc/profile.d/container_env_var.sh; '
        '[ $(id -u) -eq 0 ] && echo alias sudo="" >> ~/.bashrc;sleep infinity')
    # Use base64 to deal with the tricky quoting issues caused by runpod API.
    encoded = base64.b64encode(setup_cmd.encode('utf-8')).decode('utf-8')

    docker_args = (f'bash -c \'echo {encoded} | base64 --decode > init.sh; '
                   f'bash init.sh\'')

    # Port 8081 is occupied for nginx in the base image.
    custom_ports_str = ''
    if ports is not None:
        custom_ports_str = ''.join([f'{p}/tcp,' for p in ports])
    ports_str = (f'22/tcp,'
                 f'{custom_ports_str}'
                 f'{constants.SKY_REMOTE_RAY_DASHBOARD_PORT}/http,'
                 f'{constants.SKY_REMOTE_RAY_PORT}/http')

    image_name_formatted, template_id = _create_template_for_docker_login(
        cluster_name, image_name, docker_login_config)

    params = {
        'name': name,
        'image_name': image_name_formatted,
        'container_disk_in_gb': disk_size,
        'country_code': region,
        'data_center_id': zone,
        'ports': ports_str,
        'support_public_ip': True,
        'docker_args': docker_args,
        'template_id': template_id,
    }

    # Optional network volume mount.
    if volume_mount_path is not None:
        params['volume_mount_path'] = volume_mount_path
    if network_volume_id is not None:
        params['network_volume_id'] = network_volume_id

    # GPU instance types start with f'{gpu_count}x',
    # CPU instance types start with 'cpu'.
    is_cpu_instance = instance_type.startswith('cpu')
    if is_cpu_instance:
        # RunPod CPU instances can be uniquely identified by the instance_id.
        params.update({
            'instance_id': instance_type,
        })
    else:
        gpu_type = GPU_NAME_MAP[instance_type.split('_')[1]]
        gpu_quantity = int(instance_type.split('_')[0].replace('x', ''))
        cloud_type = instance_type.split('_')[2]
        gpu_specs = runpod.runpod.get_gpu(gpu_type)
        params.update({
            'gpu_type_id': gpu_type,
            'cloud_type': cloud_type,
            'min_vcpu_count': 4 * gpu_quantity,
            'min_memory_in_gb': gpu_specs['memoryInGb'] * gpu_quantity,
            'gpu_count': gpu_quantity,
        })

    if preemptible is None or not preemptible:
        new_instance = runpod.runpod.create_pod(**params)
    else:
        new_instance = runpod_commands.create_spot_pod(
            bid_per_gpu=bid_per_gpu,
            **params,  # type: ignore[arg-type]
        )

    return new_instance['id']


def get_registry_auth_resources(
        cluster_name: str) -> Tuple[Optional[str], Optional[str]]:
    """Gets the registry auth resources."""
    container_registry_auth_name = _construct_docker_login_template_name(
        cluster_name)
    for template in _list_pod_templates_with_container_registry():
        if template['name'] == container_registry_auth_name:
            return container_registry_auth_name, template[
                'containerRegistryAuthId']
    return None, None


def remove(instance_id: str) -> None:
    """Terminates the given instance."""
    runpod.runpod.terminate_pod(instance_id)


def get_ssh_ports(cluster_name) -> List[int]:
    """Gets the SSH ports for the given cluster."""
    logger.debug(f'Getting SSH ports for cluster {cluster_name}.')

    instances = list_instances()
    possible_names = [f'{cluster_name}-head', f'{cluster_name}-worker']

    ssh_ports = []

    for instance in instances.values():
        if instance['name'] in possible_names:
            ssh_ports.append(instance['ssh_port'])
    assert ssh_ports, (
        f'Could not find any instances for cluster {cluster_name}.')

    return ssh_ports



"""Sky subprocess daemon.
Wait for parent_pid to exit, then SIGTERM (or SIGKILL if needed) the child
processes of proc_pid.
"""
import argparse
import os
import signal
import sys
import time
from typing import List, Optional

import psutil

# Environment variable to enable kill_pg in subprocess daemon.
USE_KILL_PG_ENV_VAR = 'SKYPILOT_SUBPROCESS_DAEMON_KILL_PG'


def daemonize():
    """Detaches the process from its parent process with double-forking.

    This detachment is crucial in the context of SkyPilot and Ray job. When
    'sky cancel' is executed, it uses Ray's stop job API to terminate the job.
    Without daemonization, this subprocess_daemon process will still be a child
    of the parent process which would be terminated along with the parent
    process, ray::task or the cancel request for jobs, which is launched with
    Ray job. Daemonization ensures this process survives the 'sky cancel'
    command, allowing it to prevent orphaned processes of Ray job.
    """
    # First fork: Creates a child process identical to the parent
    if os.fork() > 0:
        # Parent process exits, allowing the child to run independently
        sys.exit()

    # Continues to run from first forked child process.
    # Detach from parent environment.
    os.setsid()

    # Second fork: Creates a grandchild process
    if os.fork() > 0:
        # First child exits, orphaning the grandchild
        sys.exit()
    # Continues execution in the grandchild process
    # This process is now fully detached from the original parent and terminal


def get_pgid_if_leader(pid) -> Optional[int]:
    """Get the process group ID of the target process if it is the leader."""
    try:
        pgid = os.getpgid(pid)
        # Only use process group if the target process is the leader. This is
        # to avoid killing the entire process group while the target process is
        # just a subprocess in the group.
        if pgid == pid:
            print(f'Process group {pgid} is the leader.')
            return pgid
        return None
    except Exception:  # pylint: disable=broad-except
        # Process group is only available in UNIX.
        return None


def kill_process_group(pgid: int) -> bool:
    """Kill the target process group."""
    try:
        print(f'Terminating process group {pgid}...')
        os.killpg(pgid, signal.SIGTERM)
    except Exception:  # pylint: disable=broad-except
        return False

    # Wait 30s for the process group to exit gracefully.
    time.sleep(30)

    try:
        print(f'Force killing process group {pgid}...')
        os.killpg(pgid, signal.SIGKILL)
    except Exception:  # pylint: disable=broad-except
        pass

    return True


def kill_process_tree(process: psutil.Process,
                      children: List[psutil.Process]) -> bool:
    """Kill the process tree of the target process."""
    if process is not None:
        # Kill the target process first to avoid having more children, or fail
        # the process due to the children being defunct.
        children = [process] + children

    if not children:
        sys.exit()

    for child in children:
        try:
            child.terminate()
        except psutil.NoSuchProcess:
            continue

    # Wait 30s for the processes to exit gracefully.
    time.sleep(30)

    # SIGKILL if they're still running.
    for child in children:
        try:
            child.kill()
        except psutil.NoSuchProcess:
            continue

    return True


def main():
    # daemonize()
    parser = argparse.ArgumentParser()
    parser.add_argument('--parent-pid', type=int, required=True)
    parser.add_argument('--proc-pid', type=int, required=True)
    parser.add_argument(
        '--initial-children',
        type=str,
        default='',
        help=(
            'Comma-separated list of initial children PIDs. This is to guard '
            'against the case where the target process has already terminated, '
            'while the children are still running.'),
    )
    args = parser.parse_args()

    process = None
    parent_process = None
    try:
        process = psutil.Process(args.proc_pid)
        parent_process = psutil.Process(args.parent_pid)
    except psutil.NoSuchProcess:
        pass

    # Initialize children list from arguments
    children = []
    if args.initial_children:
        for pid in args.initial_children.split(','):
            try:
                child = psutil.Process(int(pid))
                children.append(child)
            except (psutil.NoSuchProcess, ValueError):
                pass

    pgid: Optional[int] = None
    if os.environ.get(USE_KILL_PG_ENV_VAR) == '1':
        # Use kill_pg on UNIX system if allowed to reduce the resource usage.
        # Note that both implementations might leave subprocessed uncancelled:
        # - kill_process_tree(default): a subprocess is able to detach itself
        #   from the process tree use the same technique as daemonize(). Also,
        #   since we refresh the process tree per second, if the subprocess is
        #   launched between the [last_poll, parent_die] interval, the
        #   subprocess will not be captured will not be killed.
        # - kill_process_group: kill_pg will kill all the processed in the group
        #   but if a subprocess calls setpgid(0, 0) to detach itself from the
        #   process group (usually to daemonize itself), the subprocess will
        #   not be killed.
        pgid = get_pgid_if_leader(process.pid)

    if process is not None and parent_process is not None:
        # Wait for either parent or target process to exit
        while process.is_running() and parent_process.is_running():
            if pgid is None:
                # Refresh process tree for cleanup if process group is not
                # available.
                try:
                    tmp_children = process.children(recursive=True)
                    if tmp_children:
                        children = tmp_children
                except psutil.NoSuchProcess:
                    pass
            time.sleep(1)

    if pgid is not None:
        kill_process_group(pgid)
    else:
        kill_process_tree(process, children)


if __name__ == '__main__':
    main()



import re
from unittest import mock

import click
from click import testing as cli_testing
import pytest
import requests

from sky import clouds
from sky import exceptions
from sky import models
from sky import server
from sky.client.cli import command
from sky.schemas.api import responses
from sky.utils import status_lib
from sky.utils import ux_utils

CLOUDS_TO_TEST = [
    'aws', 'gcp', 'ibm', 'azure', 'lambda', 'scp', 'oci', 'vsphere', 'nebius'
]


def mock_server_api_version(monkeypatch, version):
    original_request = requests.request

    def mock_request(method, url, *args, **kwargs):
        if '/api/health' in url and method == 'GET':
            mock_response = mock.MagicMock()
            mock_response.status_code = 200
            mock_response.json.return_value = {
                'api_version': version,
                'version': '1.0.0',
                'commit': '1234567890'
            }
            return mock_response
        return original_request(method, url, *args, **kwargs)

    monkeypatch.setattr(requests, 'request', mock_request)


# Global API server mocking to prevent real network calls
def mock_api_server_calls(monkeypatch):
    """Mock all API server related calls to prevent real network requests."""
    # Mock API server status checks
    mock_api_info = mock.MagicMock(spec=responses.APIHealthResponse)
    mock_api_info.status = server.common.ApiServerStatus.HEALTHY
    mock_api_info.api_version = 1
    mock_api_info.version = '1.0.0'
    mock_api_info.commit = 'test_commit'
    mock_api_info.user = None
    mock_api_info.basic_auth_enabled = False

    monkeypatch.setattr('sky.server.common.get_api_server_status',
                        lambda *args, **kwargs: mock_api_info)

    # Mock other server functions that might trigger real calls
    monkeypatch.setattr('sky.server.common.check_server_healthy',
                        lambda *args, **kwargs: ('healthy', mock_api_info))
    monkeypatch.setattr('sky.server.common.get_server_url',
                        lambda *args, **kwargs: 'http://localhost:46580')
    monkeypatch.setattr('sky.server.common.is_api_server_local', lambda: True)

    # Mock make_authenticated_request to prevent real HTTP calls
    def mock_make_authenticated_request(method, path, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.history = []

        # Mock different endpoints
        if '/api/health' in path:
            mock_response.json.return_value = {
                'status': 'healthy',
                'api_version': "1",
                'version': '1.0.0',
                'version_on_disk': '1.0.0',
                'commit': 'test_commit',
                'user': None,
                'basic_auth_enabled': False
            }
        else:
            # For other endpoints, return a simple success response
            mock_response.json.return_value = {'status': 'success'}
            # Set the text attribute for request ID parsing
            mock_response.text = 'test_request_id_12345'

        return mock_response

    monkeypatch.setattr('sky.server.common.make_authenticated_request',
                        mock_make_authenticated_request)

    # Mock high-level SDK functions to prevent API calls entirely
    monkeypatch.setattr('sky.client.sdk.check',
                        lambda *args, **kwargs: 'test_request_id')

    # Create mock accelerator items that have the _asdict method like dataclasses
    class MockAcceleratorItem:

        def __init__(self,
                     accelerator_name='V100',
                     accelerator_count=1,
                     cloud='AWS',
                     instance_type='p3.2xlarge',
                     device_memory=16.0,
                     cpu_count=8,
                     memory=61.0,
                     price=1.0,
                     spot_price=0.5,
                     region='us-east-1'):
            self.accelerator_name = accelerator_name
            self.accelerator_count = accelerator_count
            self.cloud = cloud
            self.instance_type = instance_type
            self.device_memory = device_memory
            self.cpu_count = cpu_count
            self.memory = memory
            self.price = price
            self.spot_price = spot_price
            self.region = region

        def _asdict(self):
            return {
                'accelerator_name': self.accelerator_name,
                'accelerator_count': self.accelerator_count,
                'cloud': self.cloud,
                'instance_type': self.instance_type,
                'device_memory': self.device_memory,
                'cpu_count': self.cpu_count,
                'memory': self.memory,
                'price': self.price,
                'spot_price': self.spot_price,
                'region': self.region
            }

    # Mock list_accelerators to return a request ID (for specific accelerator queries)
    monkeypatch.setattr('sky.client.sdk.list_accelerators',
                        lambda *args, **kwargs: 'accelerators_request_id')

    # Mock list_accelerator_counts to return a request ID (for --all queries)
    monkeypatch.setattr('sky.client.sdk.list_accelerator_counts',
                        lambda *args, **kwargs: 'accelerator_counts_request_id')

    # Mock stream_and_get to return appropriate data based on request ID
    def mock_stream_and_get(request_id):
        if 'accelerators_request_id' in str(request_id):
            # Return dictionary with GPU names as keys and lists of accelerator items as values
            return {
                'V100': [MockAcceleratorItem('V100', 1, 'AWS', 'p3.2xlarge')],
                'T4': [MockAcceleratorItem('T4', 1, 'GCP', 'n1-standard-4')]
            }
        elif 'accelerator_counts_request_id' in str(request_id):
            # Return dictionary with GPU names as keys and lists of quantities as values
            # This is used for --all flag (when accelerator_str is None)
            return {
                'V100': [1, 2, 4, 8],  # Available quantities
                'T4': [1, 2, 4],
                'A100': [1, 2, 4, 8],
                'H100': [1, 2, 4],
            }
        elif 'status_request_id' in str(request_id):
            return [{
                'name': 'test-cluster',
                'handle': None,
                'credentials': {},
                'status': status_lib.ClusterStatus.UP,
                'workspace': 'default',
                'autostop': -1,  # -1 means disabled
                'to_down': False,
                'launched_at': 1640995200,  # Some timestamp
                'last_use': 'sky launch',
                'user_hash': 'test_user_hash',
                'user_name': 'test_user',
                'resources_str': '1x t2.micro',
                'resources_str_full': '1x t2.micro',
                'infra': 'AWS/us-east-1'
            }]
        elif 'jobs_queue_request_id' in str(request_id):
            return [{'job_id': 1, 'status': 'RUNNING'}]
        else:
            return []

    monkeypatch.setattr('sky.client.sdk.stream_and_get', mock_stream_and_get)

    # Mock sdk.get to return appropriate data based on request ID
    def mock_get(request_id):
        if 'enabled_clouds_request_id' in str(request_id):
            return ['aws', 'gcp']  # Return list of cloud names
        elif 'workspaces_request_id' in str(request_id):
            return ['default', 'workspace1']
        elif 'serve_status_request_id' in str(request_id):
            return [{
                'name': 'service1',
                'status': 'RUNNING',
                'endpoint': 'http://localhost:8080'
            }]
        else:
            return []

    monkeypatch.setattr('sky.client.sdk.get', mock_get)
    monkeypatch.setattr('sky.client.sdk.enabled_clouds',
                        lambda *args, **kwargs: 'enabled_clouds_request_id')
    monkeypatch.setattr('sky.client.sdk.realtime_kubernetes_gpu_availability',
                        lambda *args, **kwargs: 'k8s_gpu_request_id')
    monkeypatch.setattr('sky.client.sdk.kubernetes_node_info',
                        lambda *args, **kwargs: 'k8s_node_request_id')

    # Mock requests.request as well for any direct usage
    def mock_requests_request(method, url, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.text = 'test_request_id_12345'
        mock_response.json.return_value = {'status': 'success'}
        return mock_response

    monkeypatch.setattr('requests.request', mock_requests_request)
    monkeypatch.setattr('requests.get', mock_requests_request)
    monkeypatch.setattr('requests.post', mock_requests_request)

    # Mock usage tracking
    monkeypatch.setattr('sky.usage.usage_lib.messages.usage.set_internal',
                        lambda: None)

    # Mock the rest client to prevent real API calls
    def mock_rest_request(method, url, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.text = 'test_request_id_12345'
        mock_response.json.return_value = {'status': 'success'}
        return mock_response

    monkeypatch.setattr('sky.server.rest.request', mock_rest_request)
    monkeypatch.setattr('sky.server.rest.request_without_retry',
                        mock_rest_request)

    return mock_api_info


class TestWithNoCloudEnabled:

    def test_show_gpus(self, monkeypatch):
        """Tests `sky show-gpus` can be invoked (but not correctness).

        Tests below correspond to the following terminal commands, in order:

        -> sky show-gpus
        -> sky show-gpus --all
        -> sky show-gpus V100:4
        -> sky show-gpus :4
        -> sky show-gpus V100:0
        -> sky show-gpus V100:-2
        -> sky show-gpus --cloud aws --region us-west-1
        -> sky show-gpus --cloud lambda
        -> sky show-gpus --cloud lambda --all
        -> sky show-gpus V100:4 --cloud lambda
        -> sky show-gpus V100:4 --cloud lambda --all
        """
        mock_api_server_calls(monkeypatch)

        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(command.show_gpus, [])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['--all'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['V100:4'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, [':4'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['V100:0'])
        assert isinstance(result.exception, SystemExit)

        result = cli_runner.invoke(command.show_gpus, ['V100:-2'])
        assert isinstance(result.exception, SystemExit)

        result = cli_runner.invoke(command.show_gpus,
                                   ['--cloud', 'aws', '--region', 'us-west-1'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus,
                                   ['--infra', 'aws/us-west-1'])
        assert not result.exit_code

        for cloud in CLOUDS_TO_TEST:
            result = cli_runner.invoke(command.show_gpus, ['--infra', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['--cloud', cloud, '--all'])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100', '--cloud', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100:4', '--cloud', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100:4', '--cloud', cloud, '--all'])
            assert isinstance(result.exception, SystemExit)

    def test_k8s_alias_check(self, monkeypatch):
        mock_api_server_calls(monkeypatch)

        cli_runner = cli_testing.CliRunner()

        result = cli_runner.invoke(command.check, ['k8s'])
        assert not result.exit_code

        result = cli_runner.invoke(command.check, ['kubernetes'])
        assert not result.exit_code

        # With our mocking, invalid cloud names won't raise ValueError
        # since we're bypassing the real validation logic
        result = cli_runner.invoke(command.check, ['notarealcloud'])
        # In mocked environment, this should succeed rather than raise ValueError
        assert not result.exit_code

    def test_cli_api_info(self, monkeypatch):
        """Test that `sky api info` returns the expected output."""
        mock_api_info = mock_api_server_calls(monkeypatch)

        client_version = '1.0.0'
        client_commit = '1234567890'

        monkeypatch.setattr('sky.__version__', client_version)
        monkeypatch.setattr('sky.__commit__', client_commit)

        current_user = models.User(id='test_user_id', name='test_user')

        mock_get_current_user = mock.MagicMock()
        mock_get_current_user.return_value = current_user
        monkeypatch.setattr(models.User, 'get_current_user',
                            mock_get_current_user)

        user = models.User.get_current_user()
        server_url = 'http://localhost:8000'
        monkeypatch.setattr('sky.server.common.get_server_url',
                            lambda: server_url)

        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(command.api_info)
        assert not result.exit_code
        output = result.stdout.split('\n')
        assert output[
            0] == f'SkyPilot client version: {client_version}, commit: {client_commit}'
        assert output[
            1] == f'Using SkyPilot API server and dashboard: {server_url}'
        assert output[
            2] == f'├── Status: {mock_api_info.status}, commit: {mock_api_info.commit}, version: {mock_api_info.version}'
        assert output[3] == f'├── User: {current_user.name} ({current_user.id})'
        assert output[4] == f'└── Endpoint set to default local API server.'
        assert len(output) == 6


class TestHelperFunctions:

    def test_get_cluster_records_and_set_ssh_config(self, monkeypatch):
        """Tests _get_cluster_records_and_set_ssh_config with mocked components."""
        mock_api_server_calls(monkeypatch)

        # Mock cluster records that would be returned by stream_and_get
        mock_handle = mock.MagicMock()
        mock_handle.cluster_name = 'test-cluster'
        mock_handle.cached_external_ips = ['1.2.3.4']
        mock_handle.cached_external_ssh_ports = [22]
        mock_handle.docker_user = None
        mock_handle.ssh_user = 'ubuntu'
        mock_handle.launched_resources.cloud = mock.MagicMock()

        mock_records = [{
            'name': 'test-cluster',
            'handle': mock_handle,
            'credentials': {
                'ssh_user': 'ubuntu',
                'ssh_private_key': '/path/to/key.pem'
            }
        }]

        # Mock the SDK calls
        def mock_status(*args, **kwargs):
            return 'request-id'

        def mock_stream_and_get(*args, **kwargs):
            return mock_records

        monkeypatch.setattr('sky.client.sdk.status', mock_status)
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            mock_stream_and_get)

        # Mock SSHConfigHelper methods
        mock_add_cluster = mock.MagicMock()
        mock_remove_cluster = mock.MagicMock()
        mock_list_cluster_names = mock.MagicMock(
            return_value=['test-cluster', 'old-cluster'])

        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.add_cluster',
            mock_add_cluster)
        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.remove_cluster',
            mock_remove_cluster)
        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.list_cluster_names',
            mock_list_cluster_names)

        # Test case 1: Get records for specific clusters
        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records
        mock_add_cluster.assert_called_once_with('test-cluster', ['1.2.3.4'], {
            'ssh_user': 'ubuntu',
            'ssh_private_key': '/path/to/key.pem'
        }, [22], None, 'ubuntu')
        # Shouldn't remove anything because all clusters provided are in the returned records
        mock_remove_cluster.assert_not_called()

        # Reset mocks for next test
        mock_add_cluster.reset_mock()
        mock_remove_cluster.reset_mock()

        # Test case 2: Get records for all users
        records = command._get_cluster_records_and_set_ssh_config(
            None, all_users=True)
        assert records == mock_records
        mock_add_cluster.assert_called_once()
        # Should remove old-cluster since it's not in the returned records
        mock_remove_cluster.assert_called_once_with('old-cluster')

        # Test case 3: Test with a cluster that has no handle/credentials
        mock_records_no_handle = [{'name': 'test-cluster', 'handle': None}]
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            lambda *args, **kwargs: mock_records_no_handle)

        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records_no_handle
        # Should remove the cluster from SSH config since it has no handle
        mock_remove_cluster.assert_called_with('test-cluster')

        # Reset mocks for next test
        mock_add_cluster.reset_mock()
        mock_remove_cluster.reset_mock()

        # Test case 4: Test with a cluster that is using kubernetes
        mock_k8s_handle = mock.MagicMock()
        mock_k8s_handle.cluster_name = 'test-cluster'
        mock_k8s_handle.cached_external_ips = ['1.2.3.4']
        mock_k8s_handle.cached_external_ssh_ports = [22]
        mock_k8s_handle.docker_user = None
        mock_k8s_handle.ssh_user = 'ubuntu'
        mock_k8s_handle.launched_resources.cloud = clouds.Kubernetes()
        mock_records_kubernetes = [{
            'name': 'test-cluster',
            'handle': mock_k8s_handle,
            'credentials': {
                'ssh_user': 'ubuntu',
                'ssh_private_key': '/path/to/key.pem'
            },
        }]
        server_url = 'http://localhost:8000'
        monkeypatch.setattr('sky.server.common.get_server_url',
                            lambda: server_url)
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            lambda *args, **kwargs: mock_records_kubernetes)
        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records_kubernetes
        mock_add_cluster.assert_called_once()
        added_cluster_args = mock_add_cluster.call_args
        assert added_cluster_args[0][0] == 'test-cluster'
        assert added_cluster_args[0][1] == ['1.2.3.4']
        # proxy command should be set, but is dependent on the server url, so we don't check the exact value
        assert added_cluster_args[0][2].get('ssh_proxy_command') is not None
        assert server_url in added_cluster_args[0][2].get('ssh_proxy_command')
        assert added_cluster_args[0][2].get(
            'ssh_private_key') == '/path/to/key.pem'
        assert added_cluster_args[0][2].get('ssh_user') == 'ubuntu'
        assert added_cluster_args[0][3] == [22]
        assert added_cluster_args[0][4] is None
        assert added_cluster_args[0][5] == 'ubuntu'
        mock_remove_cluster.assert_not_called()

    def test_list_to_str_float_formatting(self):
        """Test that _list_to_str formats whole number floats as integers.
        
        Regression test for GitHub issue #6484 where requestable quantities
        were shown as '1.0, 2.0, 4.0, 8.0' instead of '1, 2, 4, 8'.
        """

        # Define the function locally to test it (since it's nested in show_gpus)
        def _list_to_str(lst):

            def format_number(e):
                # If it's a float that's a whole number, display as int
                if isinstance(e, float) and e.is_integer():
                    return str(int(e))
                return str(e)

            return ', '.join([format_number(e) for e in lst])

        # Test case from GitHub issue #6484: whole number floats should display as integers
        assert _list_to_str([1.0, 2.0, 4.0, 8.0]) == '1, 2, 4, 8'

        # Test mixed floats and integers
        assert _list_to_str([1, 2.0, 4.0, 8]) == '1, 2, 4, 8'

        # Test fractional floats should remain as floats
        assert _list_to_str([1.5, 2.0, 4.0]) == '1.5, 2, 4'

        # Test edge cases
        assert _list_to_str([0.0, 1.0]) == '0, 1'
        assert _list_to_str([0.5, 1.0, 2.5]) == '0.5, 1, 2.5'

        # Test integers remain unchanged
        assert _list_to_str([1, 2, 4, 8]) == '1, 2, 4, 8'

        # Test empty list
        assert _list_to_str([]) == ''

        # Test single element
        assert _list_to_str([1.0]) == '1'
        assert _list_to_str([1.5]) == '1.5'

    def test_show_gpus_k8s_float_formatting(self, monkeypatch):
        """Integration test for sky show-gpus --infra k8s output formatting.
        
        Regression test for GitHub issue #6484 to ensure that requestable quantities
        are displayed as integers (1, 2, 4, 8) instead of floats (1.0, 2.0, 4.0, 8.0).
        """
        mock_api_server_calls(monkeypatch)

        # Import required modules
        from sky import models

        # Mock kubernetes GPU availability data with float counts (the bug scenario)
        mock_gpu_availability = [
            models.RealtimeGpuAvailability(
                gpu='H100_NVLINK_80GB',
                counts=[1.0, 2.0, 4.0,
                        8.0],  # These should be formatted as integers
                capacity=16,
                available=16)
        ]

        # Mock enabled clouds to include kubernetes
        def mock_get(request_id):
            if 'enabled_clouds_request_id' in str(request_id):
                return ['kubernetes']  # Enable kubernetes
            return []

        monkeypatch.setattr('sky.client.sdk.get', mock_get)
        monkeypatch.setattr('sky.client.sdk.enabled_clouds',
                            lambda *args, **kwargs: 'enabled_clouds_request_id')

        # Mock the kubernetes contexts and GPU availability
        monkeypatch.setattr('sky.clouds.Kubernetes.existing_allowed_contexts',
                            lambda: ['test-context'])
        monkeypatch.setattr(
            'sky.client.sdk.realtime_kubernetes_gpu_availability',
            lambda *args, **kwargs: 'k8s_gpu_request_id')
        monkeypatch.setattr('sky.client.sdk.kubernetes_node_info',
                            lambda *args, **kwargs: 'k8s_node_request_id')

        # Mock node info to avoid additional API calls
        mock_node_info = models.KubernetesNodesInfo(node_info_dict={},
                                                    hint=None)

        def mock_stream_and_get_conditional(request_id):
            if 'k8s_gpu' in request_id:
                return [('test-context', mock_gpu_availability)]
            elif 'k8s_node' in request_id:
                return mock_node_info
            return []

        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            mock_stream_and_get_conditional)

        # Run the command
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(command.show_gpus, ['--infra', 'k8s'])

        # Check that command succeeded
        assert result.exit_code == 0, f"Command failed with output: {result.output}"

        # Check that the output contains properly formatted integers (not floats)
        output = result.output
        assert '1, 2, 4, 8' in output, f"Expected '1, 2, 4, 8' in output, got: {output}"
        # Ensure it doesn't contain the problematic float format
        assert '1.0, 2.0, 4.0, 8.0' not in output, f"Found float format in output: {output}"


def strip_ansi(s: str) -> str:
    return re.sub(r"\x1b\[[0-9;]*m", "", s)


class DummyCloudError(exceptions.CloudError):

    def __init__(self):
        super().__init__(
            message="Request error UNAUTHENTICATED: Invalid token",
            cloud_provider="nebius",
            error_type="RequestError",
        )


@pytest.mark.parametrize("mode", ["down", "stop", "autostop"])
def test_batch_continues_on_errors_helper(monkeypatch, capsys, mode):
    monkeypatch.setenv("RICH_FORCE_TERMINAL", "1")
    monkeypatch.setenv("RICH_PROGRESS_NO_CLEAR", "1")
    monkeypatch.setenv("TERM", "xterm-256color")
    monkeypatch.setenv("COLUMNS", "100")

    names = ["sky-ok-1", "sky-nebius-fail", "sky-ok-2"]

    def fake_down(name, purge=False):
        if name == "sky-nebius-fail":
            raise DummyCloudError()

    def fake_stop(name, purge=False):
        return fake_down(name, purge=purge)

    def fake_autostop(name, idle_minutes, wait_for, down):
        return fake_down(name)

    monkeypatch.setattr(command, '_async_call_or_wait',
                        lambda *args, **kwargs: None)
    monkeypatch.setattr(command.sdk, "down", fake_down)
    monkeypatch.setattr(command.sdk, "stop", fake_stop)
    monkeypatch.setattr(command.sdk, "autostop", fake_autostop)

    monkeypatch.setattr(command,
                        "_get_cluster_records_and_set_ssh_config",
                        lambda clusters=None, all_users=False: [{
                            "name": n,
                            "status": None
                        } for n in (clusters or names)])

    class FakeControllers:

        @staticmethod
        def from_name(name, expect_exact_match):
            return None

    monkeypatch.setattr(command, "controller_utils",
                        type("X", (), {"Controllers": FakeControllers}))

    kwargs = dict(
        names=names,
        apply_to_all=False,
        all_users=False,
        down=(mode == "down"),
        no_confirm=True,
        purge=False,
        idle_minutes_to_autostop=(10 if mode == "autostop" else None),
        wait_for=None,
        async_call=False,
    )

    monkeypatch.setattr(
        command.sdk, "get", lambda *args, **kwargs: [{
            "name": n,
            "status": None
        } for n in names])

    command._down_or_stop_clusters(**kwargs)

    captured = capsys.readouterr()

    out_raw = (captured.out + captured.err).replace("\r", "\n")
    out = strip_ansi(out_raw)

    assert re.search(r"nebius.*UNAUTHENTICATED", out, re.I)
    assert "Summary:" in out
    assert "✗ Failed: sky-nebius-fail" in out
    assert "Failed:" in out

    if mode == "down":
        assert "Terminating cluster sky-ok-1...done" in out
        assert "Terminating cluster sky-ok-2...done" in out
    elif mode == "stop":
        assert "Stopping cluster sky-ok-1...done" in out
        assert "Stopping cluster sky-ok-2...done" in out
    else:  # autostop
        assert "Scheduling autostop on cluster 'sky-ok-1'...done" in out
        assert "Scheduling autostop on cluster 'sky-ok-2'...done" in out

    assert "✓ Succeeded:" in out
    summary_line = next(line for line in out.splitlines()
                        if line.strip().startswith("✓ Succeeded:"))
    succ_list = [n.strip() for n in summary_line.split(":", 1)[1].split(",")]
    assert set(succ_list) == {"sky-ok-1", "sky-ok-2"}


"""Utilities for formatting tables for CLI output."""
import abc
from datetime import datetime
from typing import Any, Dict, List, Optional

import prettytable

from sky import sky_logging
from sky.jobs import utils as managed_jobs
from sky.schemas.api import responses
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import log_utils
from sky.utils import volume

logger = sky_logging.init_logger(__name__)


def format_job_queue(jobs: List[responses.ClusterJobRecord]):
    """Format the job queue for display.

    Usage:
        jobs = get_job_queue()
        print(format_job_queue(jobs))
    """
    job_table = log_utils.create_table([
        'ID', 'NAME', 'USER', 'SUBMITTED', 'STARTED', 'DURATION', 'RESOURCES',
        'STATUS', 'LOG', 'GIT COMMIT'
    ])
    for job in jobs:
        job_table.add_row([
            job.job_id,
            job.job_name,
            job.username,
            log_utils.readable_time_duration(job.submitted_at),
            log_utils.readable_time_duration(job.start_at),
            log_utils.readable_time_duration(job.start_at,
                                             job.end_at,
                                             absolute=True),
            job.resources,
            job.status.colored_str(),
            job.log_path,
            job.metadata.get('git_commit', '-'),
        ])
    return job_table


def format_storage_table(storages: List[responses.StorageRecord],
                         show_all: bool = False) -> str:
    """Format the storage table for display.

    Args:
        storage_table (dict): The storage table.

    Returns:
        str: The formatted storage table.
    """
    storage_table = log_utils.create_table([
        'NAME',
        'UPDATED',
        'STORE',
        'COMMAND',
        'STATUS',
    ])

    for row in storages:
        launched_at = row.launched_at
        if show_all:
            command = row.last_use
        else:
            command = common_utils.truncate_long_string(
                row.last_use, constants.LAST_USE_TRUNC_LENGTH)
        storage_table.add_row([
            # NAME
            row.name,
            # LAUNCHED
            log_utils.readable_time_duration(launched_at),
            # CLOUDS
            ', '.join([s.value for s in row.store]),
            # COMMAND,
            command,
            # STATUS
            row.status.value,
        ])
    if storages:
        return str(storage_table)
    else:
        return 'No existing storage.'


def format_job_table(
    jobs: List[responses.ManagedJobRecord],
    show_all: bool,
    show_user: bool,
    pool_status: Optional[List[Dict[str, Any]]] = None,
    max_jobs: Optional[int] = None,
    status_counts: Optional[Dict[str, int]] = None,
):
    jobs = [job.model_dump() for job in jobs]
    return managed_jobs.format_job_table(
        jobs,
        pool_status=pool_status,
        show_all=show_all,
        show_user=show_user,
        max_jobs=max_jobs,
        job_status_counts=status_counts,
    )


_BASIC_COLUMNS = [
    'NAME',
    'TYPE',
    'INFRA',
    'SIZE',
    'USER',
    'WORKSPACE',
    'AGE',
    'STATUS',
    'LAST_USE',
    'USED_BY',
]


def _get_infra_str(cloud: Optional[str], region: Optional[str],
                   zone: Optional[str]) -> str:
    """Get the infrastructure string for the volume."""
    infra = ''
    if cloud:
        infra += cloud
    if region:
        infra += f'/{region}'
    if zone:
        infra += f'/{zone}'
    return infra


class VolumeTable(abc.ABC):
    """The volume table."""

    def __init__(self,
                 volumes: List[responses.VolumeRecord],
                 show_all: bool = False):
        super().__init__()
        self.table = self._create_table(show_all)
        self._add_rows(volumes, show_all)

    def _get_row_base_columns(self,
                              row: responses.VolumeRecord,
                              show_all: bool = False) -> List[str]:
        """Get the base columns for a row."""
        # Convert last_attached_at timestamp to human readable string
        last_attached_at = row.get('last_attached_at')
        if last_attached_at is not None:
            last_attached_at_str = datetime.fromtimestamp(
                last_attached_at).strftime('%Y-%m-%d %H:%M:%S')
        else:
            last_attached_at_str = '-'
        size = row.get('size', '')
        if size:
            size = f'{size}Gi'
        usedby_str = '-'
        usedby_clusters = row.get('usedby_clusters')
        usedby_pods = row.get('usedby_pods')
        if usedby_clusters:
            usedby_str = f'{", ".join(usedby_clusters)}'
        elif usedby_pods:
            usedby_str = f'{", ".join(usedby_pods)}'
        if show_all:
            usedby = usedby_str
        else:
            usedby = common_utils.truncate_long_string(
                usedby_str, constants.USED_BY_TRUNC_LENGTH)
        infra = _get_infra_str(row.get('cloud'), row.get('region'),
                               row.get('zone'))
        return [
            row.get('name', ''),
            row.get('type', ''),
            infra,
            size,
            row.get('user_name', '-'),
            row.get('workspace', '-'),
            log_utils.human_duration(row.get('launched_at', 0)),
            row.get('status', ''),
            last_attached_at_str,
            usedby,
        ]

    def _create_table(self, show_all: bool = False) -> prettytable.PrettyTable:
        """Create the volume table."""
        raise NotImplementedError

    def _add_rows(self,
                  volumes: List[responses.VolumeRecord],
                  show_all: bool = False) -> None:
        """Add rows to the volume table."""
        raise NotImplementedError

    @abc.abstractmethod
    def format(self) -> str:
        """Format the volume table for display."""
        raise NotImplementedError


class PVCVolumeTable(VolumeTable):
    """The PVC volume table."""

    def _create_table(self, show_all: bool = False) -> prettytable.PrettyTable:
        """Create the PVC volume table."""
        #  If show_all is False, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY
        #  If show_all is True, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY, NAME_ON_CLOUD
        #   STORAGE_CLASS, ACCESS_MODE

        if show_all:
            columns = _BASIC_COLUMNS + [
                'NAME_ON_CLOUD',
                'STORAGE_CLASS',
                'ACCESS_MODE',
            ]
        else:
            columns = _BASIC_COLUMNS

        table = log_utils.create_table(columns)
        return table

    def _add_rows(self,
                  volumes: List[responses.VolumeRecord],
                  show_all: bool = False) -> None:
        """Add rows to the PVC volume table."""
        for row in volumes:
            table_row = self._get_row_base_columns(row, show_all)
            if show_all:
                table_row.append(row.get('name_on_cloud', ''))
                table_row.append(
                    row.get('config', {}).get('storage_class_name', '-'))
                table_row.append(row.get('config', {}).get('access_mode', ''))

            self.table.add_row(table_row)

    def format(self) -> str:
        """Format the PVC volume table for display."""
        return 'Kubernetes PVCs:\n' + str(self.table)


class RunPodVolumeTable(VolumeTable):
    """The RunPod volume table."""

    def _create_table(self, show_all: bool = False) -> prettytable.PrettyTable:
        """Create the RunPod volume table."""
        #  If show_all is False, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY
        #  If show_all is True, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY, NAME_ON_CLOUD

        if show_all:
            columns = _BASIC_COLUMNS + ['NAME_ON_CLOUD']
        else:
            columns = _BASIC_COLUMNS

        table = log_utils.create_table(columns)
        return table

    def _add_rows(self,
                  volumes: List[responses.VolumeRecord],
                  show_all: bool = False) -> None:
        """Add rows to the RunPod volume table."""
        for row in volumes:
            table_row = self._get_row_base_columns(row, show_all)
            if show_all:
                table_row.append(row.get('name_on_cloud', ''))

            self.table.add_row(table_row)

    def format(self) -> str:
        """Format the RunPod volume table for display."""
        return 'RunPod Network Volumes:\n' + str(self.table)


def format_volume_table(volumes: List[responses.VolumeRecord],
                        show_all: bool = False) -> str:
    """Format the volume table for display.

    Args:
        volume_table (dict): The volume table.

    Returns:
        str: The formatted volume table.
    """
    volumes_per_type: Dict[str, List[responses.VolumeRecord]] = {}
    supported_volume_types = [
        volume_type.value for volume_type in volume.VolumeType
    ]
    for row in volumes:
        volume_type = row.get('type', '')
        if volume_type in supported_volume_types:
            if volume_type not in volumes_per_type:
                volumes_per_type[volume_type] = []
            volumes_per_type[volume_type].append(row)
        else:
            logger.warning(f'Unknown volume type: {volume_type}')
            continue
    table_str = ''
    for volume_type, volume_list in volumes_per_type.items():
        if table_str:
            table_str += '\n\n'
        if volume_type == volume.VolumeType.PVC.value:
            pvc_table = PVCVolumeTable(volume_list, show_all)
            table_str += pvc_table.format()
        elif volume_type == volume.VolumeType.RUNPOD_NETWORK_VOLUME.value:
            runpod_table = RunPodVolumeTable(volume_list, show_all)
            table_str += runpod_table.format()
    if table_str:
        return table_str
    else:
        return 'No existing volumes.'



"""Utility functions for benchmarking."""

import functools
import logging
import time
from typing import Callable, Optional

from sky import sky_logging

logger = sky_logging.init_logger(__name__)


def log_execution_time(func: Optional[Callable] = None,
                       *,
                       name: Optional[str] = None,
                       level: int = logging.DEBUG,
                       precision: int = 4) -> Callable:
    """Mark a function and log its execution time.

    Args:
        func: Function to decorate.
        name: Name of the function.
        level: Logging level.
        precision: Number of decimal places (default: 4).

    Usage:
        from sky.utils import benchmark_utils

        @benchmark_utils.log_execution_time
        def my_function():
            pass

        @benchmark_utils.log_execution_time(name='my_module.my_function2')
        def my_function2():
            pass
    """

    def decorator(f: Callable) -> Callable:

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            nonlocal name
            name = name or f.__name__
            start_time = time.perf_counter()
            try:
                result = f(*args, **kwargs)
                return result
            finally:
                end_time = time.perf_counter()
                execution_time = end_time - start_time
                log = (f'Method {name} executed in '
                       f'{execution_time:.{precision}f}')
                logger.log(level, log)

        return wrapper

    if func is None:
        return decorator
    else:
        return decorator(func)



"""SCP Open-API Functions.

This module contains a set of rest api functions accessing SCP Open-API.
"""
import base64
import datetime
from functools import wraps
import hashlib
import hmac
import json
import logging
import os
import time
import typing
from typing import Any, Dict, List, Optional
from urllib import parse

from sky.adaptors import common as adaptors_common

if typing.TYPE_CHECKING:
    import requests
else:
    requests = adaptors_common.LazyImport('requests')

CREDENTIALS_PATH = '~/.scp/scp_credential'
API_ENDPOINT = 'https://openapi.samsungsdscloud.com'
TEMP_VM_JSON_PATH = '/tmp/json/tmp_vm_body.json'

logger = logging.getLogger(__name__)


class SCPClientError(Exception):
    pass


class SCPOngoingRequestError(Exception):
    pass


class SCPCreationFailError(Exception):
    pass


class Metadata:
    """Per-cluster metadata file."""

    def __init__(self, path_prefix: str, cluster_name: str) -> None:
        # TODO(ewzeng): Metadata file is not thread safe. This is fine for
        # now since SkyPilot uses a per-cluster lock for ray-related
        # operations. In the future, add a filelock around __getitem__,
        # __setitem__ and refresh.
        self.path = os.path.expanduser(f'{path_prefix}-{cluster_name}')
        # In case parent directory does not exist
        os.makedirs(os.path.dirname(self.path), exist_ok=True)

    def __getitem__(self, instance_id: str) -> Dict[str, Any]:
        assert os.path.exists(self.path), 'Metadata file not found'
        with open(self.path, 'r', encoding='utf-8') as f:
            metadata = json.load(f)
        return metadata.get(instance_id)

    def __setitem__(self, instance_id: str, value: Optional[Dict[str,
                                                                 Any]]) -> None:
        # Read from metadata file
        if os.path.exists(self.path):
            with open(self.path, 'r', encoding='utf-8') as f:
                metadata = json.load(f)
        else:
            metadata = {}
        # Update metadata
        if value is None:
            if instance_id in metadata:
                metadata.pop(instance_id)  # del entry
            if not metadata:
                if os.path.exists(self.path):
                    os.remove(self.path)
                return
        else:
            metadata[instance_id] = value
        # Write to metadata file
        with open(self.path, 'w', encoding='utf-8') as f:
            json.dump(metadata, f)

    def refresh(self, instance_ids: List[str]) -> None:
        """Remove all tags for instances not in instance_ids."""
        if not os.path.exists(self.path):
            return
        with open(self.path, 'r', encoding='utf-8') as f:
            metadata = json.load(f)
        for instance_id in list(metadata.keys()):
            if instance_id not in instance_ids:
                del metadata[instance_id]
        if not metadata:
            os.remove(self.path)
            return
        with open(self.path, 'w', encoding='utf-8') as f:
            json.dump(metadata, f)

    def keys(self):
        if not os.path.exists(self.path):
            return []
        with open(self.path, 'r', encoding='utf-8') as f:
            metadata = json.load(f)
            return list(metadata.keys())


def raise_scp_error(response: 'requests.Response') -> None:
    """Raise SCPCloudError if appropriate. """
    status_code = response.status_code
    if status_code in (200, 202):
        return
    try:
        resp_json = response.json()
        message = resp_json['message']
    except (KeyError, json.decoder.JSONDecodeError) as e:
        raise SCPClientError('Unexpected error. Status code: '
                             f'{status_code}') from e

    if status_code == 404:
        raise SCPCreationFailError(f'{status_code}: {message}')

    if 'There is an ongoing request' in message:
        raise SCPOngoingRequestError(f'{status_code}: {message}')

    raise SCPClientError(f'{status_code}: {message}')


def _singleton(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


def _retry(method, max_tries=50, backoff_s=5):

    @wraps(method)
    def method_with_retries(self, *args, **kwargs):
        try_count = 0
        while try_count < max_tries:
            try:
                return method(self, *args, **kwargs)
            except SCPOngoingRequestError:
                logger.warning('Caught a Ongoing Request. Retrying.')
                try_count += 1
                if try_count < max_tries:
                    time.sleep(backoff_s)
                else:
                    raise

    return method_with_retries


@_singleton
class SCPClient:
    """Wrapper functions for SCP Cloud API."""

    def __init__(self) -> None:
        self.credentials = os.path.expanduser(CREDENTIALS_PATH)
        assert os.path.exists(self.credentials), 'Credentials not found'
        with open(self.credentials, 'r', encoding='utf-8') as f:
            lines = [line.strip() for line in f.readlines() if ' = ' in line]
            self._credentials = {
                line.split(' = ')[0]: line.split(' = ')[1] for line in lines
            }
        self.access_key = self._credentials['access_key']
        self.secret_key = self._credentials['secret_key']
        self.project_id = self._credentials['project_id']
        self.client_type = 'OpenApi'
        self.timestamp = ''
        self.signature = ''

        self.headers = {
            'X-Cmp-AccessKey': f'{self.access_key}',
            'X-Cmp-ClientType': f'{self.client_type}',
            'X-Cmp-ProjectId': f'{self.project_id}',
            'X-Cmp-Timestamp': f'{self.timestamp}',
            'X-Cmp-Signature': f'{self.signature}'
        }

    def create_instance(self, instance_config):
        """Launch new instances."""
        url = f'{API_ENDPOINT}/virtual-server/v4/virtual-servers'
        return self._post(url, instance_config)

    @_retry
    def _get(self, url, contents_key='contents'):
        method = 'GET'
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.get(url, headers=self.headers)
        raise_scp_error(response)
        if contents_key is not None:
            return response.json().get(contents_key, [])
        else:
            return response.json()

    @_retry
    def _post(self, url, request_body):
        method = 'POST'
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.post(url, json=request_body, headers=self.headers)

        raise_scp_error(response)
        return response.json()

    @_retry
    def _delete(self, url, request_body=None):
        method = 'DELETE'
        self.set_timestamp()
        self.set_signature(url=url, method=method)
        if request_body:
            response = requests.delete(url,
                                       json=request_body,
                                       headers=self.headers)

        else:
            response = requests.delete(url, headers=self.headers)
        raise_scp_error(response)
        return response.json()

    def create_security_group(self, zone_id, vpc_id, sg_name):
        url = f'{API_ENDPOINT}/security-group/v3/security-groups'
        request_body = {
            'loggable': False,
            'securityGroupName': sg_name,
            'serviceZoneId': zone_id,
            'vpcId': vpc_id,
            'securityGroupDescription': 'sky security group'
        }
        return self._post(url, request_body)

    def _security_group_rule_not_exist(self, sg_id, direction, ports):
        response = self.get_security_group_rules(sg_id)
        rules = []
        for rule in response:
            rule_direction = rule['ruleDirection']
            if rule_direction == direction:
                rules.append(rule)
        for rule in rules:
            port_list = ','.join(rule['tcpServices'])
            port = ','.join(ports)
            if port == port_list:
                return False
        return True

    def add_security_group_rule(self, sg_id, direction,
                                ports: Optional[List[str]]):
        if ports is None:
            if direction == 'IN':
                ports = ['22']
            else:
                ports = ['21', '22', '80', '443']
        services = []
        for port in ports:
            services.append({'serviceType': 'TCP', 'serviceValue': port})
        if self._security_group_rule_not_exist(sg_id, direction, ports):
            url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}/rules'  # pylint: disable=line-too-long
            if direction == 'IN':
                target_address = 'sourceIpAddresses'
            else:
                target_address = 'destinationIpAddresses'
            request_body = {
                'ruleDirection': direction,
                'services': services,
                target_address: ['0.0.0.0/0'],
                'ruleDescription': 'sky security group rule'
            }
            return self._post(url, request_body)

    def _firewall_rule_not_exist(self, firewall_id, internal_ip, direction,
                                 ports):
        response = self.get_firewall_rules(firewall_id)
        rules = []
        for rule in response:
            if direction == 'IN':
                if internal_ip == rule['destinationIpAddresses'][0]:
                    rules.append(rule)
            else:
                if internal_ip == rule['sourceIpAddresses'][0]:
                    rules.append(rule)
        for rule in rules:
            port_list = ','.join(rule['tcpServices'])
            port = ','.join(ports)
            if port == port_list:
                return False
        return True

    def add_firewall_rule(self, firewall_id, internal_ip, direction, ports):
        if ports is None:
            if direction == 'IN':
                ports = ['22']
            else:
                ports = ['21', '22', '80', '443']
        services = []
        for port in ports:
            services.append({'serviceType': 'TCP', 'serviceValue': port})
        if self._firewall_rule_not_exist(firewall_id, internal_ip, direction,
                                         ports):
            url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
            if direction == 'IN':
                source_ip = '0.0.0.0/0'
                destination_ip = internal_ip
            else:
                source_ip = internal_ip
                destination_ip = '0.0.0.0/0'
            request_body = {
                'sourceIpAddresses': [source_ip],
                'destinationIpAddresses': [destination_ip],
                'services': services,
                'ruleDirection': direction,
                'ruleAction': 'ALLOW',
                'isRuleEnabled': True,
                'ruleLocationType': 'FIRST',
                'ruleDescription': 'sky firewall rule'
            }
            return self._post(url, request_body)

    def terminate_instance(self, instance_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{instance_id}'
        return self._delete(url)

    def get_instances(self) -> List[dict]:
        """List existing instances."""
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers'
        return self._get(url)

    def get_catalog(self) -> Dict[str, Any]:
        """List offered instances and their availability."""
        response = requests.get(f'{API_ENDPOINT}/instance-types',
                                headers=self.headers)
        raise_scp_error(response)
        return response.json().get('data', [])

    def get_signature(self, method: str, url: str) -> str:
        url_info = parse.urlsplit(url)
        url = (f'{url_info.scheme}://{url_info.netloc}'
               f'{parse.quote(url_info.path)}')
        if url_info.query:
            enc_params = list(
                map(lambda item: (item[0], parse.quote(item[1][0])),
                    parse.parse_qs(url_info.query).items()))
            url = f'{url}?{parse.urlencode(enc_params)}'

        message = method + url + self.timestamp \
                  + self.access_key + self.project_id + self.client_type
        message = bytes(message, 'utf-8')
        secret = bytes(self.secret_key, 'utf-8')
        signature = str(
            base64.b64encode(
                hmac.new(secret, message, digestmod=hashlib.sha256).digest()),
            'utf-8')

        return str(signature)

    def set_timestamp(self) -> None:
        self.timestamp = str(
            int(
                round(
                    datetime.datetime.timestamp(datetime.datetime.now() -
                                                datetime.timedelta(minutes=1)) *
                    1000)))
        self.headers['X-Cmp-Timestamp'] = self.timestamp

    def set_signature(self, method: str, url: str) -> None:
        self.signature = self.get_signature(url=url, method=method)
        self.headers['X-Cmp-Signature'] = self.signature

    def get_nic(self, instance_id) -> List[dict]:
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{instance_id}/nics'  # pylint: disable=line-too-long
        return self._get(url)

    def get_external_ip(self, instance_id, ip):
        nics = self.get_nic(instance_id=instance_id)
        for nic in nics:
            if (nic['ip'] == ip and nic['subnetType'] == 'PUBLIC'):
                return nic['natIp']
        return None

    def get_zones(self) -> List[dict]:
        url = f'{API_ENDPOINT}/project/v3/projects/{self.project_id}/zones'
        return self._get(url)

    def get_vpcs(self, service_zone_id) -> List[dict]:
        url = f'{API_ENDPOINT}/vpc/v2/vpcs?serviceZoneId={service_zone_id}'
        return self._get(url)

    def get_subnets(self) -> List[dict]:
        url = f'{API_ENDPOINT}/subnet/v2/subnets?subnetTypes=PUBLIC'
        return self._get(url)

    def delete_security_group(self, sg_id):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}'
        return self._delete(url)

    def delete_firewall_rule(self, firewall_id, rule_ids):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
        request_body = {'ruleDeletionType': 'PARTIAL', 'ruleIds': rule_ids}
        return self._delete(url, request_body=request_body)

    def get_security_groups(self, vpc_id=None, sg_name=None):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups'
        parameter = []
        if vpc_id is not None:
            parameter.append('vpcId=' + vpc_id)
        if sg_name is not None:
            parameter.append('securityGroupName=' + sg_name)
        if len(parameter) > 0:
            url = url + '?' + '&'.join(parameter)
        return self._get(url)

    def get_internet_gateway(self):
        url = f'{API_ENDPOINT}/internet-gateway/v2/internet-gateways'
        return self._get(url)

    def get_firewall_rule_info(self, firewall_id, rule_id):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules/{rule_id}'  # pylint: disable=line-too-long
        return self._get(url, contents_key=None)

    def get_firewalls(self):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls'
        return self._get(url)

    def get_service_zone_names(self):
        url = f'{API_ENDPOINT}/project/v3/projects/{self.project_id}/zones'
        zone_contents = self._get(url)
        return [content['serviceZoneName'] for content in zone_contents]

    def start_instance(self, instance_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{instance_id}/start'  # pylint: disable=line-too-long
        return self._post(url=url, request_body={})

    def stop_instance(self, instance_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{instance_id}/stop'  # pylint: disable=line-too-long
        return self._post(url=url, request_body={})

    def get_security_group_rules(self, sg_id):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}/rules'
        return self._get(url)

    def get_firewall_rules(self, firewall_id):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
        return self._get(url)

    def get_instance_info(self, instance_id):
        url = f'{API_ENDPOINT}/virtual-server/v3/virtual-servers/{instance_id}'
        return self._get(url=url, contents_key=None)

    def create_vpc(self, zone_id):
        vpc_name = 'skyvpc' + zone_id[5:10]
        request_body = {
            'serviceZoneId': zone_id,
            'vpcName': vpc_name,
            'vpcDescription': 'sky vpc'
        }
        url = f'{API_ENDPOINT}/vpc/v3/vpcs'
        return self._post(url, request_body)

    def create_subnet(self, vpc_id, zone_id):
        subnet_name = 'skysubnet' + zone_id[5:10]
        request_body = {
            'subnetCidrBlock': '192.168.0.0/24',
            'subnetName': subnet_name,
            'subnetType': 'PUBLIC',
            'vpcId': vpc_id,
            'subnetDescription': 'sky subnet'
        }
        url = f'{API_ENDPOINT}/subnet/v2/subnets'
        return self._post(url, request_body)

    def create_internet_gateway(self, vpc_id):
        request_body = {
            'firewallEnabled': True,
            'firewallLoggable': False,
            'internetGatewayType': 'SHARED',
            'vpcId': vpc_id,
            'internetGatewayDescription': 'sky internet gateway'
        }
        url = f'{API_ENDPOINT}/internet-gateway/v4/internet-gateways'
        return self._post(url, request_body)

    def get_vpc_info(self, vpc_id):
        url = f'{API_ENDPOINT}/vpc/v2/vpcs/{vpc_id}'
        return self._get(url=url, contents_key=None)

    def get_subnet_info(self, subnet_id):
        url = f'{API_ENDPOINT}/subnet/v2/subnets/{subnet_id}'
        return self._get(url=url, contents_key=None)

    def get_internet_gateway_info(self, internet_gateway_id):
        url = f'{API_ENDPOINT}/internet-gateway/v2/internet-gateways/{internet_gateway_id}'  # pylint: disable=line-too-long
        return self._get(url=url, contents_key=None)

    def get_key_pairs(self):
        url = f'{API_ENDPOINT}/key-pair/v1/key-pairs'
        return self._get(url=url, contents_key=None)



"""Temporary storage context manager."""

import contextlib
import contextvars
import functools
import os
import tempfile
import typing
from typing import Any, Callable, Iterator, Optional, TypeVar

_TEMP_DIR: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    'temp_store_dir', default=None)


@contextlib.contextmanager
def tempdir() -> Iterator[str]:
    """Context manager for temporary directory of current context.

    This wraps tempfile.TemporaryDirectory and makes the temp dir available
    throughout the context, eliminating the need to pass the temp dir to
    the nested functions that need it.

    This context manager is nestable - nested calls will create new temp dirs
    and restore the previous temp dir when exiting.
    """
    with tempfile.TemporaryDirectory(prefix='sky-tmp') as temp_dir:
        token = _TEMP_DIR.set(temp_dir)
        try:
            yield temp_dir
        finally:
            _TEMP_DIR.reset(token)


# Keep the function signature same as tempfile.mkdtemp.
# pylint: disable=redefined-builtin
def mkdtemp(suffix: Optional[str] = None,
            prefix: Optional[str] = None,
            dir: Optional[str] = None) -> str:
    """Create a temporary directory in the temp dir of current context.

    The directory will be cleaned when the current context exits.
    If there is no temp dir in current context, this function is equivalent to
    tempfile.mkdtemp.
    """
    context_temp_dir = _TEMP_DIR.get()

    if context_temp_dir is not None and dir is None:
        dir = context_temp_dir
    elif context_temp_dir is not None and dir is not None:
        dir = os.path.join(context_temp_dir, dir)
        os.makedirs(dir, exist_ok=True)

    return tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)


F = TypeVar('F', bound=Callable[..., Any])


def with_tempdir(func: F) -> F:
    """Decorator that wraps a function call with tempdir() context manager.

    Refer to `tempdir` for more details.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with tempdir():
            return func(*args, **kwargs)

    return typing.cast(F, wrapper)



"""This module provides functions to generate GraphQL mutations for deploying
spot instance Pods on RunPod.

Reference:
    https://github.com/runpod/runpod-python/blob/main/runpod/api/ctl_commands.py

Functions:
    generate_spot_pod_deployment_mutation: Generates a GraphQL mutation string
        for deploying a spot instance Pod on RunPod.

Example:
    >>> mutation = generate_spot_pod_deployment_mutation(
            name='test',
            image_name='runpod/stack',
            gpu_type_id='NVIDIA GeForce RTX 3070',
            bid_per_gpu=0.3
        )
"""
from typing import List, Optional

from sky.adaptors import runpod
from sky.provision.runpod.api.pods import generate_spot_pod_deployment_mutation

_INTERRUPTABLE_POD_FIELD: str = 'podRentInterruptable'
_RESPONSE_DATA_FIELD: str = 'data'


def create_spot_pod(
    name: str,
    image_name: str,
    gpu_type_id: str,
    bid_per_gpu: float,
    cloud_type: str = 'ALL',
    volume_mount_path: str = '/runpod-volume',
    gpu_count: Optional[int] = 1,
    min_memory_in_gb: Optional[int] = 1,
    min_vcpu_count: Optional[int] = 1,
    container_disk_in_gb: Optional[int] = None,
    volume_in_gb: Optional[int] = 0,
    ports: Optional[str] = None,
    start_ssh: Optional[bool] = True,
    start_jupyter: Optional[bool] = False,
    env: Optional[dict] = None,
    docker_args: Optional[str] = '',
    support_public_ip: Optional[bool] = True,
    terminate_after: Optional[str] = None,
    stop_after: Optional[str] = None,
    data_center_id: Optional[str] = None,
    country_code: Optional[str] = None,
    network_volume_id: Optional[str] = None,
    allowed_cuda_versions: Optional[List[str]] = None,
    min_download: Optional[int] = None,
    min_upload: Optional[int] = None,
    cuda_version: Optional[str] = None,
    template_id: Optional[str] = None,
    volume_key: Optional[str] = None,
) -> dict:
    """This module provides functions to generate GraphQL mutations for
    deploying spot instance Pods on RunPod.

    Functions:
        generate_spot_pod_deployment_mutation: Generates a GraphQL mutation
            string for deploying a spot instance Pod on RunPod.

    Example:
        >>> mutation = generate_spot_pod_deployment_mutation(
                name='test',
                image_name='runpod/stack',
                gpu_type_id='NVIDIA GeForce RTX 3070',
                bid_per_gpu=0.3
            )
    """
    runpod.runpod.get_gpu(gpu_type_id)
    # refer to https://graphql-spec.runpod.io/#definition-CloudTypeEnum
    if cloud_type not in ['ALL', 'COMMUNITY', 'SECURE']:
        raise ValueError('cloud_type must be one of ALL, COMMUNITY or SECURE')

    if network_volume_id and data_center_id is None:
        user_info = runpod.runpod.get_user()
        for network_volume in user_info['networkVolumes']:
            if network_volume['id'] == network_volume_id:
                data_center_id = network_volume['dataCenterId']
                break

    if container_disk_in_gb is None and template_id is None:
        container_disk_in_gb = 10

    mutation = generate_spot_pod_deployment_mutation(
        name=name,
        image_name=image_name,
        gpu_type_id=gpu_type_id,
        bid_per_gpu=bid_per_gpu,
        cloud_type=cloud_type,
        gpu_count=gpu_count,
        min_memory_in_gb=min_memory_in_gb,
        min_vcpu_count=min_vcpu_count,
        container_disk_in_gb=container_disk_in_gb,
        volume_in_gb=volume_in_gb,
        volume_mount_path=volume_mount_path,
        ports=ports,
        start_ssh=start_ssh,
        start_jupyter=start_jupyter,
        env=env,
        docker_args=docker_args,
        support_public_ip=support_public_ip,
        terminate_after=terminate_after,
        stop_after=stop_after,
        data_center_id=data_center_id,
        country_code=country_code,
        network_volume_id=network_volume_id,
        allowed_cuda_versions=allowed_cuda_versions,
        min_download=min_download,
        min_upload=min_upload,
        cuda_version=cuda_version,
        template_id=template_id,
        volume_key=volume_key,
    )
    response = runpod.runpod.api.graphql.run_graphql_query(mutation)
    return response[_RESPONSE_DATA_FIELD][_INTERRUPTABLE_POD_FIELD]



"""CoreWeave cloud adaptor."""

import configparser
import contextlib
import os
import threading
from typing import Dict, Optional, Tuple

from sky import exceptions
from sky import sky_logging
from sky.adaptors import common
from sky.clouds import cloud
from sky.utils import annotations
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

COREWEAVE_PROFILE_NAME = 'cw'
COREWEAVE_CREDENTIALS_PATH = '~/.coreweave/cw.credentials'
COREWEAVE_CONFIG_PATH = '~/.coreweave/cw.config'
NAME = 'CoreWeave'
DEFAULT_REGION = 'US-EAST-01A'
_DEFAULT_ENDPOINT = 'https://cwobject.com'
_INDENT_PREFIX = '    '

_IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for CoreWeave.'
                         'Try pip install "skypilot[coreweave]"')

boto3 = common.LazyImport('boto3', import_error_message=_IMPORT_ERROR_MESSAGE)
botocore = common.LazyImport('botocore',
                             import_error_message=_IMPORT_ERROR_MESSAGE)

_LAZY_MODULES = (boto3, botocore)
_session_creation_lock = threading.RLock()


@contextlib.contextmanager
def _load_cw_credentials_env():
    """Context manager to temporarily change the AWS credentials file path."""
    prev_credentials_path = os.environ.get('AWS_SHARED_CREDENTIALS_FILE')
    prev_config_path = os.environ.get('AWS_CONFIG_FILE')
    os.environ['AWS_SHARED_CREDENTIALS_FILE'] = COREWEAVE_CREDENTIALS_PATH
    os.environ['AWS_CONFIG_FILE'] = COREWEAVE_CONFIG_PATH
    try:
        yield
    finally:
        if prev_credentials_path is None:
            del os.environ['AWS_SHARED_CREDENTIALS_FILE']
        else:
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = prev_credentials_path
        if prev_config_path is None:
            del os.environ['AWS_CONFIG_FILE']
        else:
            os.environ['AWS_CONFIG_FILE'] = prev_config_path


def get_coreweave_credentials(boto3_session):
    """Gets the CoreWeave credentials from the boto3 session object.

    Args:
        boto3_session: The boto3 session object.
    Returns:
        botocore.credentials.ReadOnlyCredentials object with the CoreWeave
        credentials.
    """
    with _load_cw_credentials_env():
        coreweave_credentials = boto3_session.get_credentials()
        if coreweave_credentials is None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('CoreWeave credentials not found. Run '
                                 '`sky check` to verify credentials are '
                                 'correctly set up.')
        return coreweave_credentials.get_frozen_credentials()


@annotations.lru_cache(scope='global')
def session():
    """Create an AWS session for CoreWeave."""
    # Creating the session object is not thread-safe for boto3,
    # so we add a reentrant lock to synchronize the session creation.
    # Reference: https://github.com/boto/boto3/issues/1592
    # However, the session object itself is thread-safe, so we are
    # able to use lru_cache() to cache the session object.
    with _session_creation_lock:
        with _load_cw_credentials_env():
            session_ = boto3.session.Session(
                profile_name=COREWEAVE_PROFILE_NAME)
        return session_


@annotations.lru_cache(scope='global')
def resource(resource_name: str, **kwargs):
    """Create a CoreWeave resource.

    Args:
        resource_name: CoreWeave resource name (e.g., 's3').
        kwargs: Other options.
    """
    # Need to use the resource retrieved from the per-thread session
    # to avoid thread-safety issues (Directly creating the client
    # with boto3.resource() is not thread-safe).
    # Reference: https://stackoverflow.com/a/59635814

    session_ = session()
    coreweave_credentials = get_coreweave_credentials(session_)
    endpoint = get_endpoint()

    return session_.resource(
        resource_name,
        endpoint_url=endpoint,
        aws_access_key_id=coreweave_credentials.access_key,
        aws_secret_access_key=coreweave_credentials.secret_key,
        region_name='auto',
        config=botocore.config.Config(s3={'addressing_style': 'virtual'}),
        **kwargs)


@annotations.lru_cache(scope='global')
def client(service_name: str):
    """Create CoreWeave client of a certain service.

    Args:
        service_name: CoreWeave service name (e.g., 's3').
    """
    # Need to use the client retrieved from the per-thread session
    # to avoid thread-safety issues (Directly creating the client
    # with boto3.client() is not thread-safe).
    # Reference: https://stackoverflow.com/a/59635814

    session_ = session()
    coreweave_credentials = get_coreweave_credentials(session_)
    endpoint = get_endpoint()

    return session_.client(
        service_name,
        endpoint_url=endpoint,
        aws_access_key_id=coreweave_credentials.access_key,
        aws_secret_access_key=coreweave_credentials.secret_key,
        region_name='auto',
        config=botocore.config.Config(s3={'addressing_style': 'virtual'}),
    )


@common.load_lazy_modules(_LAZY_MODULES)
def botocore_exceptions():
    """AWS botocore exception."""
    # pylint: disable=import-outside-toplevel
    from botocore import exceptions as boto_exceptions
    return boto_exceptions


def get_endpoint():
    """Parse the COREWEAVE_CONFIG_PATH to get the endpoint_url.

    The config file is an AWS-style config file with format:
        [profile cw]
        endpoint_url = https://cwobject.com
        s3 =
            addressing_style = virtual

    Returns:
        str: The endpoint URL from the config file, or the default endpoint
             if the file doesn't exist or doesn't contain the endpoint_url.
    """
    config_path = os.path.expanduser(COREWEAVE_CONFIG_PATH)
    if not os.path.isfile(config_path):
        return _DEFAULT_ENDPOINT

    try:
        config = configparser.ConfigParser()
        config.read(config_path)

        # Try to get endpoint_url from [profile cw] section
        profile_section = f'profile {COREWEAVE_PROFILE_NAME}'
        if config.has_section(profile_section):
            if config.has_option(profile_section, 'endpoint_url'):
                endpoint = config.get(profile_section, 'endpoint_url')
                return endpoint.strip()
    except (configparser.Error, OSError) as e:
        logger.warning(f'Failed to parse CoreWeave config file: {e}. '
                       f'Using default endpoint: {_DEFAULT_ENDPOINT}')

    return _DEFAULT_ENDPOINT


def check_credentials(
        cloud_capability: cloud.CloudCapability) -> Tuple[bool, Optional[str]]:
    if cloud_capability == cloud.CloudCapability.STORAGE:
        return check_storage_credentials()
    else:
        raise exceptions.NotSupportedError(
            f'{NAME} does not support {cloud_capability}.')


def check_storage_credentials() -> Tuple[bool, Optional[str]]:
    """Checks if the user has access credentials to CoreWeave Object Storage.

    Returns:
        A tuple of a boolean value and a hint message where the bool
        is True when both credentials needed for CoreWeave storage is set.
        It is False when either of those are not set, which would hint with a
        string on unset credential.
    """
    hints = None
    profile_in_cred = coreweave_profile_in_cred()
    profile_in_config = coreweave_profile_in_config()

    if not profile_in_cred:
        hints = (f'[{COREWEAVE_PROFILE_NAME}] profile is not set in '
                 f'{COREWEAVE_CREDENTIALS_PATH}.')
    if not profile_in_config:
        if hints:
            hints += ' Additionally, '
        else:
            hints = ''
        hints += (f'[{COREWEAVE_PROFILE_NAME}] profile is not set in '
                  f'{COREWEAVE_CONFIG_PATH}.')

    if hints:
        hints += ' Run the following commands:'
        if not profile_in_cred:
            hints += f'\n{_INDENT_PREFIX}  $ pip install boto3'
            hints += (f'\n{_INDENT_PREFIX}  $ AWS_SHARED_CREDENTIALS_FILE='
                      f'{COREWEAVE_CREDENTIALS_PATH} aws configure --profile '
                      f'{COREWEAVE_PROFILE_NAME}')
        if not profile_in_config:
            hints += (f'\n{_INDENT_PREFIX}  $ AWS_CONFIG_FILE='
                      f'{COREWEAVE_CONFIG_PATH} aws configure set endpoint_url'
                      f' <ENDPOINT_URL> --profile '
                      f'{COREWEAVE_PROFILE_NAME}')
            hints += (f'\n{_INDENT_PREFIX}  $ AWS_CONFIG_FILE='
                      f'{COREWEAVE_CONFIG_PATH} aws configure set '
                      f's3.addressing_style virtual --profile '
                      f'{COREWEAVE_PROFILE_NAME}')
        hints += f'\n{_INDENT_PREFIX}For more info: '
        hints += 'https://docs.coreweave.com/docs/products/storage/object-storage/get-started-caios'  # pylint: disable=line-too-long

    return (False, hints) if hints else (True, hints)


def coreweave_profile_in_config() -> bool:
    """Checks if CoreWeave profile is set in config"""
    conf_path = os.path.expanduser(COREWEAVE_CONFIG_PATH)
    coreweave_profile_exists = False
    if os.path.isfile(conf_path):
        with open(conf_path, 'r', encoding='utf-8') as file:
            for line in file:
                if f'[profile {COREWEAVE_PROFILE_NAME}]' in line:
                    coreweave_profile_exists = True
                    break
    return coreweave_profile_exists


def coreweave_profile_in_cred() -> bool:
    """Checks if CoreWeave profile is set in credentials"""
    cred_path = os.path.expanduser(COREWEAVE_CREDENTIALS_PATH)
    coreweave_profile_exists = False
    if os.path.isfile(cred_path):
        with open(cred_path, 'r', encoding='utf-8') as file:
            for line in file:
                if f'[{COREWEAVE_PROFILE_NAME}]' in line:
                    coreweave_profile_exists = True
                    break
    return coreweave_profile_exists


def get_credential_file_mounts() -> Dict[str, str]:
    """Returns credential file mounts for CoreWeave.

    Returns:
        Dict[str, str]: A dictionary mapping source paths to destination paths
        for credential files.
    """
    coreweave_credential_mounts = {
        COREWEAVE_CREDENTIALS_PATH: COREWEAVE_CREDENTIALS_PATH,
        COREWEAVE_CONFIG_PATH: COREWEAVE_CONFIG_PATH
    }
    return coreweave_credential_mounts



"""Azure configuration bootstrapping.

Creates the resource group and deploys the configuration template to Azure for
a cluster to be launched.
"""
import hashlib
import json
from pathlib import Path
import random
import time
from typing import Any, Callable, Tuple

from sky import exceptions
from sky import sky_logging
from sky.adaptors import azure
from sky.provision import common
from sky.provision import constants
from sky.utils import common_utils

logger = sky_logging.init_logger(__name__)

UNIQUE_ID_LEN = 4
_RESOURCE_GROUP_WAIT_FOR_DELETION_TIMEOUT = 480  # 8 minutes
_CLUSTER_ID = '{cluster_name_on_cloud}-{unique_id}'


def get_azure_sdk_function(client: Any, function_name: str) -> Callable:
    """Retrieve a callable function from Azure SDK client object.

    Newer versions of the various client SDKs renamed function names to
    have a begin_ prefix. This function supports both the old and new
    versions of the SDK by first trying the old name and falling back to
    the prefixed new name.
    """
    func = getattr(client, function_name,
                   getattr(client, f'begin_{function_name}', None))
    if func is None:
        raise AttributeError(
            f'{client.__name__!r} object has no {function_name} or '
            f'begin_{function_name} attribute')
    return func


def get_cluster_id_and_nsg_name(resource_group: str,
                                cluster_name_on_cloud: str) -> Tuple[str, str]:
    hasher = hashlib.md5(resource_group.encode('utf-8'))
    unique_id = hasher.hexdigest()[:UNIQUE_ID_LEN]
    # We use the cluster name + resource group hash as the
    # unique ID for the cluster, as we need to make sure that
    # the deployments have unique names during failover.
    cluster_id = _CLUSTER_ID.format(cluster_name_on_cloud=cluster_name_on_cloud,
                                    unique_id=unique_id)
    nsg_name = f'sky-{cluster_id}-nsg'
    return cluster_id, nsg_name


@common.log_function_start_end
def bootstrap_instances(
        region: str, cluster_name_on_cloud: str,
        config: common.ProvisionConfig) -> common.ProvisionConfig:
    """See sky/provision/__init__.py"""
    # TODO: use new azure sdk instead of ARM deployment.
    del region  # unused
    provider_config = config.provider_config
    subscription_id = provider_config.get('subscription_id')
    if subscription_id is None:
        subscription_id = azure.get_subscription_id()
    # Increase the timeout to fix the Azure get-access-token (used by ray azure
    # node_provider) timeout issue.
    # Tracked in https://github.com/Azure/azure-cli/issues/20404#issuecomment-1249575110 # pylint: disable=line-too-long
    resource_client = azure.get_client('resource', subscription_id)
    provider_config['subscription_id'] = subscription_id
    logger.info(f'Using subscription id: {subscription_id}')

    assert (
        'resource_group'
        in provider_config), 'Provider config must include resource_group field'
    resource_group = provider_config['resource_group']

    assert ('location'
            in provider_config), 'Provider config must include location field'
    params = {'location': provider_config['location']}

    assert ('use_external_resource_group'
            in provider_config), ('Provider config must include '
                                  'use_external_resource_group field')
    use_external_resource_group = provider_config['use_external_resource_group']

    if 'tags' in provider_config:
        params['tags'] = provider_config['tags']

    # When resource group is user specified, it already exists in certain
    # region.
    if not use_external_resource_group:
        logger.info(f'Creating/Updating resource group: {resource_group}')
        rg_create_or_update = get_azure_sdk_function(
            client=resource_client.resource_groups,
            function_name='create_or_update')
        rg_creation_start = time.time()
        retry = 0
        while (time.time() - rg_creation_start <
               _RESOURCE_GROUP_WAIT_FOR_DELETION_TIMEOUT):
            try:
                rg_create_or_update(resource_group_name=resource_group,
                                    parameters=params)
                break
            except azure.exceptions().ResourceExistsError as e:
                if 'ResourceGroupBeingDeleted' in str(e):
                    if retry % 5 == 0:
                        logger.info(
                            f'Azure resource group {resource_group} of a '
                            'recent terminated cluster '
                            f'{cluster_name_on_cloud} is being deleted. It can'
                            ' only be provisioned after it is fully deleted. '
                            'Waiting...')
                    time.sleep(1)
                    retry += 1
                    continue
                raise
            except azure.exceptions().ClientAuthenticationError as e:
                message = (
                    'Failed to authenticate with Azure. Please check your '
                    'Azure credentials. ClientAuthenticationError: '
                    f'{common_utils.format_exception(e)}').replace('\n', ' ')
                logger.error(message)
                raise exceptions.NoClusterLaunchedError(message) from e
        else:
            message = (
                f'Timed out waiting for resource group {resource_group} to be '
                'deleted.')
            logger.error(message)
            raise TimeoutError(message)

    # load the template file
    current_path = Path(__file__).parent
    template_path = current_path.joinpath('azure-config-template.json')
    with open(template_path, 'r', encoding='utf-8') as template_fp:
        template = json.load(template_fp)

    logger.info(f'Using cluster name: {cluster_name_on_cloud}')

    cluster_id, nsg_name = get_cluster_id_and_nsg_name(
        resource_group=provider_config['resource_group'],
        cluster_name_on_cloud=cluster_name_on_cloud)
    subnet_mask = provider_config.get('subnet_mask')
    if subnet_mask is None:
        # choose a random subnet, skipping most common value of 0
        random.seed(cluster_id)
        subnet_mask = f'10.{random.randint(1, 254)}.0.0/16'
    logger.info(f'Using subnet mask: {subnet_mask}')

    parameters = {
        'properties': {
            'mode': azure.deployment_mode().incremental,
            'template': template,
            'parameters': {
                'subnet': {
                    'value': subnet_mask
                },
                'clusterId': {
                    'value': cluster_id
                },
                'nsgName': {
                    'value': nsg_name
                },
                'location': {
                    'value': params['location']
                }
            },
        }
    }

    # Skip creating or updating the deployment if the deployment already exists
    # and the cluster name is the same.
    get_deployment = get_azure_sdk_function(client=resource_client.deployments,
                                            function_name='get')
    deployment_exists = False
    if use_external_resource_group:
        deployment_name = (
            constants.EXTERNAL_RG_BOOTSTRAP_DEPLOYMENT_NAME.format(
                cluster_name_on_cloud=cluster_name_on_cloud))
        deployment_list = [deployment_name]
    else:
        deployment_name = constants.DEPLOYMENT_NAME
        deployment_list = [
            constants.DEPLOYMENT_NAME, constants.LEGACY_DEPLOYMENT_NAME
        ]

    for deploy_name in deployment_list:
        try:
            deployment = get_deployment(resource_group_name=resource_group,
                                        deployment_name=deploy_name)
            logger.info(f'Deployment {deploy_name!r} already exists. '
                        'Skipping deployment creation.')

            outputs = deployment.properties.outputs
            if outputs is not None:
                deployment_exists = True
                break
        except azure.exceptions().ResourceNotFoundError:
            deployment_exists = False

    if not deployment_exists:
        logger.info(f'Creating/Updating deployment: {deployment_name}')
        create_or_update = get_azure_sdk_function(
            client=resource_client.deployments,
            function_name='create_or_update')
        # TODO (skypilot): this takes a long time (> 40 seconds) to run.
        outputs = create_or_update(
            resource_group_name=resource_group,
            deployment_name=deployment_name,
            parameters=parameters,
        ).result().properties.outputs

    # append output resource ids to be used with vm creation
    provider_config['msi'] = outputs['msi']['value']
    provider_config['nsg'] = outputs['nsg']['value']
    provider_config['subnet'] = outputs['subnet']['value']

    return config



"""Utilities for managing managed job file content.

The helpers in this module fetch job file content (DAG YAML/env files) from the
database-first storage added for managed jobs, transparently falling back to
legacy on-disk paths when needed. Consumers should prefer the string-based
helpers so controllers never have to rely on local disk state.
"""

import os
from typing import Optional

from sky import sky_logging
from sky.jobs import state as managed_job_state

logger = sky_logging.init_logger(__name__)


def get_job_dag_content(job_id: int) -> Optional[str]:
    """Get DAG YAML content for a job from database or disk.

    Args:
        job_id: The job ID

    Returns:
        DAG YAML content as string, or None if not found
    """
    file_info = managed_job_state.get_job_file_contents(job_id)

    # Prefer content stored in the database
    if file_info['dag_yaml_content'] is not None:
        return file_info['dag_yaml_content']

    # Fallback to disk path for backward compatibility
    dag_yaml_path = file_info.get('dag_yaml_path')
    if dag_yaml_path and os.path.exists(dag_yaml_path):
        try:
            with open(dag_yaml_path, 'r', encoding='utf-8') as f:
                content = f.read()
                logger.debug('Loaded DAG YAML from disk for job %s: %s', job_id,
                             dag_yaml_path)
                return content
        except (FileNotFoundError, IOError, OSError) as e:
            logger.warning(
                f'Failed to read DAG YAML from disk {dag_yaml_path}: {e}')

    logger.warning(f'DAG YAML content not found for job {job_id}')
    return None


def get_job_env_content(job_id: int) -> Optional[str]:
    """Get environment file content for a job from database or disk.

    Args:
        job_id: The job ID

    Returns:
        Environment file content as string, or None if not found
    """
    file_info = managed_job_state.get_job_file_contents(job_id)

    # Prefer content stored in the database
    if file_info['env_file_content'] is not None:
        return file_info['env_file_content']

    # Fallback to disk path for backward compatibility
    env_file_path = file_info.get('env_file_path')
    if env_file_path and os.path.exists(env_file_path):
        try:
            with open(env_file_path, 'r', encoding='utf-8') as f:
                content = f.read()
                logger.debug('Loaded environment file from disk for job %s: %s',
                             job_id, env_file_path)
                return content
        except (FileNotFoundError, IOError, OSError) as e:
            logger.warning(
                f'Failed to read environment file from disk {env_file_path}: '
                f'{e}')

    # Environment file is optional, so don't warn if not found
    return None



"""Analyze the new catalog fetched with the original."""
import typing
from typing import List

from sky.adaptors import common as adaptors_common
from sky.catalog import common

if typing.TYPE_CHECKING:
    import pandas as pd
else:
    pd = adaptors_common.LazyImport('pandas')


def resource_diff(original_df: 'pd.DataFrame', new_df: 'pd.DataFrame',
                  check_tuple: List[str]) -> 'pd.DataFrame':
    """Returns the difference between two dataframes."""
    original_resources = original_df[check_tuple]
    new_resources = new_df[check_tuple]

    return new_resources.merge(
        original_resources, on=check_tuple, how='left',
        indicator=True)[lambda x: x['_merge'] == 'left_only'].sort_values(
            by=check_tuple)


CLOUD_CHECKS = {
    'aws': ['InstanceType', 'Region', 'AvailabilityZone'],
    'azure': ['InstanceType', 'Region'],
    'gcp': ['InstanceType', 'Region', 'AcceleratorName', 'AcceleratorCount']
}

table = {}

for cloud, current_check_tuple in CLOUD_CHECKS.items():
    result = {}
    print(f'=> Checking {cloud}')
    original_catalog_df = common.read_catalog(f'{cloud}.csv')
    new_catalog_df = pd.read_csv(f'{cloud}.csv')

    resource_diff_df = resource_diff(original_catalog_df, new_catalog_df,
                                     current_check_tuple)
    resource_diff_df.merge(new_catalog_df, on=current_check_tuple,
                           how='left').to_csv(f'{cloud}_diff.csv', index=False)

    result['#resources'] = len(resource_diff_df)

    check_price = current_check_tuple + ['Price']
    price_diff_df = resource_diff(original_catalog_df, new_catalog_df,
                                  check_price)
    result['#prices'] = len(price_diff_df)

    check_price = current_check_tuple + ['SpotPrice']
    spot_price_diff_df = resource_diff(original_catalog_df, new_catalog_df,
                                       check_price)
    result['#spot_prices'] = len(spot_price_diff_df)

    table[cloud] = result

summary = pd.DataFrame(table).T
summary.to_csv('diff_summary.csv')
print(summary)



"""Responses for the API server."""

import enum
from typing import Any, Dict, List, Optional

import pydantic

from sky import data
from sky import models
from sky.jobs import state as job_state
from sky.server import common
from sky.skylet import job_lib
from sky.utils import status_lib


class ResponseBaseModel(pydantic.BaseModel):
    """A pydantic model that acts like a dict.

    Supports the following syntax:
    class SampleResponse(DictLikePayload):
        field: str

    response = SampleResponse(field='value')
    print(response['field']) # prints 'value'
    response['field'] = 'value2'
    print(response['field']) # prints 'value2'
    print('field' in response) # prints True

    This model exists for backwards compatibility with the
    old SDK that used to return a dict.

    The backward compatibility may be removed
    in the future.
    """
    # Ignore extra fields in the request body, which is useful for backward
    # compatibility. The difference with `allow` is that `ignore` will not
    # include the unknown fields when dump the model, i.e., we can add new
    # fields to the request body without breaking the existing old API server
    # where the handler function does not accept the new field in function
    # signature.
    model_config = pydantic.ConfigDict(extra='ignore')

    # backward compatibility with dict
    # TODO(syang): remove this in v0.13.0
    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError as e:
            raise KeyError(key) from e

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __contains__(self, key):
        return hasattr(self, key)

    def keys(self):
        return self.model_dump().keys()

    def values(self):
        return self.model_dump().values()

    def items(self):
        return self.model_dump().items()

    def __repr__(self):
        return self.__dict__.__repr__()


class APIHealthResponse(ResponseBaseModel):
    """Response for the API health endpoint."""
    status: common.ApiServerStatus
    api_version: str = ''
    version: str = ''
    version_on_disk: str = ''
    commit: str = ''
    # Whether basic auth on api server is enabled
    basic_auth_enabled: bool = False
    user: Optional[models.User] = None
    # Whether service account token is enabled
    service_account_token_enabled: bool = False
    # Whether basic auth on ingress is enabled
    ingress_basic_auth_enabled: bool = False


class StatusResponse(ResponseBaseModel):
    """Response for the status endpoint."""
    name: str
    launched_at: int
    # pydantic cannot generate the pydantic-core schema for
    # backends.ResourceHandle, so we use Any here.
    # This is an internally facing field anyway, so it's less
    # of a problem that it's not typed.
    handle: Optional[Any] = None
    last_use: Optional[str] = None
    status: status_lib.ClusterStatus
    autostop: int
    to_down: bool
    owner: Optional[List[str]] = None
    # metadata is a JSON, so we use Any here.
    metadata: Optional[Dict[str, Any]] = None
    cluster_hash: str
    cluster_ever_up: bool
    status_updated_at: Optional[int] = None
    user_hash: str
    user_name: str
    config_hash: Optional[str] = None
    workspace: str
    last_creation_yaml: Optional[str] = None
    last_creation_command: Optional[str] = None
    is_managed: bool
    last_event: Optional[str] = None
    resources_str: Optional[str] = None
    resources_str_full: Optional[str] = None
    # credentials is a JSON, so we use Any here.
    credentials: Optional[Dict[str, Any]] = None
    nodes: int
    cloud: Optional[str] = None
    region: Optional[str] = None
    cpus: Optional[str] = None
    memory: Optional[str] = None
    accelerators: Optional[str] = None
    cluster_name_on_cloud: Optional[str] = None


class ClusterJobRecord(ResponseBaseModel):
    """Response for the cluster job queue endpoint."""
    job_id: int
    job_name: str
    username: str
    user_hash: str
    submitted_at: float
    # None if the job has not started yet.
    start_at: Optional[float] = None
    # None if the job has not ended yet.
    end_at: Optional[float] = None
    resources: str
    status: job_lib.JobStatus
    log_path: str
    metadata: Dict[str, Any] = {}


class UploadStatus(enum.Enum):
    """Status of the upload."""
    UPLOADING = 'uploading'
    COMPLETED = 'completed'


class StorageRecord(ResponseBaseModel):
    """Response for the storage list endpoint."""
    name: str
    launched_at: int
    store: List[data.StoreType]
    last_use: str
    status: status_lib.StorageStatus


# TODO (syang) figure out which fields are always present
# and therefore can be non-optional.
class ManagedJobRecord(ResponseBaseModel):
    """A single managed job record."""
    # The job_id in the spot table
    task_job_id: Optional[int] = pydantic.Field(None, alias='_job_id')
    job_id: Optional[int] = None
    task_id: Optional[int] = None
    job_name: Optional[str] = None
    task_name: Optional[str] = None
    job_duration: Optional[float] = None
    workspace: Optional[str] = None
    status: Optional[job_state.ManagedJobStatus] = None
    schedule_state: Optional[str] = None
    resources: Optional[str] = None
    cluster_resources: Optional[str] = None
    cluster_resources_full: Optional[str] = None
    cloud: Optional[str] = None
    region: Optional[str] = None
    zone: Optional[str] = None
    infra: Optional[str] = None
    recovery_count: Optional[int] = None
    details: Optional[str] = None
    failure_reason: Optional[str] = None
    user_name: Optional[str] = None
    user_hash: Optional[str] = None
    submitted_at: Optional[float] = None
    start_at: Optional[float] = None
    end_at: Optional[float] = None
    user_yaml: Optional[str] = None
    entrypoint: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    controller_pid: Optional[int] = None
    dag_yaml_path: Optional[str] = None
    env_file_path: Optional[str] = None
    last_recovered_at: Optional[float] = None
    run_timestamp: Optional[str] = None
    priority: Optional[int] = None
    original_user_yaml_path: Optional[str] = None
    pool: Optional[str] = None
    pool_hash: Optional[str] = None
    current_cluster_name: Optional[str] = None
    job_id_on_pool_cluster: Optional[int] = None
    accelerators: Optional[Dict[str, int]] = None


class VolumeRecord(ResponseBaseModel):
    """A single volume record."""
    name: str
    type: str
    launched_at: int
    cloud: str
    region: Optional[str] = None
    zone: Optional[str] = None
    size: Optional[str] = None
    config: Dict[str, Any]
    name_on_cloud: str
    user_hash: str
    user_name: str
    workspace: str
    last_attached_at: Optional[int] = None
    last_use: Optional[str] = None
    status: Optional[str] = None
    usedby_pods: List[str]
    usedby_clusters: List[str]



"""VIM Utils
"""

import re
import subprocess
import time
from typing import List

from sky import sky_logging
from sky.adaptors import vsphere as vsphere_adaptor
from sky.catalog.data_fetchers.fetch_vsphere import get_accelerators_from_csv

logger = sky_logging.init_logger(__name__)
DISPLAY_CONTROLLER_CLASS_ID_PREFIXES = ['03']
VMWARE_VIRTUAL_DISPLAY_CONTROLLER_IDS = ['0000:00:0f.0']


def get_objs_by_names(content, vimtype: type, names: List[str]):
    """    Get the vsphere managed object associated with a given text name
    """
    # Create a set for the names for faster lookups
    names_set = set(names)
    objs = {}
    container = content.viewManager.CreateContainerView(content.rootFolder,
                                                        vimtype, True)
    for entity in container.view:
        if entity.name in names_set:
            objs[entity.name] = entity
            names_set.remove(entity.name)

    container.Destroy()

    # If there are any names left in the names_set, those were not found
    if names_set:
        for name in names_set:
            logger.error(f'Object \'{name}\' not found')
        raise RuntimeError(f'Only {len(objs)} out of {len(names)} '
                           f'objects found')

    return objs


def get_hosts_by_cluster_names(content, vcenter_name, cluster_name_dicts=None):
    """    Get a list of hosts in the vCenter Server by cluster names
    """
    if cluster_name_dicts is None:
        cluster_name_dicts = [{}]
    hosts = []
    gpu_list = get_accelerators_from_csv()
    if not cluster_name_dicts:
        cluster_view = content.viewManager.CreateContainerView(
            content.rootFolder,
            [vsphere_adaptor.get_vim().ClusterComputeResource], True)
        cluster_name_dicts = [{
            'name': cluster.name
        } for cluster in cluster_view.view]
        cluster_view.Destroy()
        if not cluster_name_dicts:
            logger.warning(f'vCenter \'{vcenter_name}\' has no clusters')

    # Retrieve all cluster names from the cluster_name_dicts
    cluster_names = [
        cluster_dict['name'] for cluster_dict in cluster_name_dicts
    ]

    # Use get_objs to retrieve all cluster objects at once
    cluster_obj_dict = get_objs_by_names(
        content, [vsphere_adaptor.get_vim().ClusterComputeResource],
        cluster_names)

    for cluster_name_dict in cluster_name_dicts:
        cluster_name = cluster_name_dict.get('name')
        cluster_obj = cluster_obj_dict.get(cluster_name)
        if cluster_obj:
            datacenter_name = get_datacenter_name(cluster_obj)
            hosts_info = list_hosts_with_devices_info(cluster_obj.host,
                                                      vcenter_name,
                                                      datacenter_name,
                                                      cluster_name, gpu_list)
            hosts.extend(hosts_info)
        else:
            logger.warning(f'Cluster \'{cluster_name}\' not found')
    return hosts


def get_datacenter_name(cluster_obj):
    datacenter_name = None
    parent = cluster_obj.parent
    while parent:
        if isinstance(parent, vsphere_adaptor.get_vim().Datacenter):
            datacenter_name = parent.name
            break
        parent = parent.parent
    return datacenter_name


def filter_vms_by_cluster_name(content, cluster_names=None):
    """    Get a list of vms in the vCenter clusters
    If cluster_names is empty or None, then get all VMs in vCenter
    """
    instances = []
    container = content.viewManager.CreateContainerView(
        content.rootFolder, [vsphere_adaptor.get_vim().ClusterComputeResource],
        True)
    cluster_objs = container.view
    container.Destroy()
    # Filter the target cluster, and get the VMs in cluster
    for cluster in cluster_objs:
        if cluster_names and cluster.name not in cluster_names:
            continue
        for host in cluster.host:
            vms = host.vm
            instances.extend(vms)
    return instances


def list_hosts_with_devices_info(hosts, vcenter_name, datacenter_name,
                                 cluster_name, gpu_list):

    def decimal_to_four_char_hex(decimal):
        result = str(hex(decimal)).replace('0x', '').zfill(4)
        return f'{result}'

    all_hosts_with_devices = []

    # Get the host information
    for host in hosts:
        hw = host.hardware
        system_info = hw.systemInfo
        cpu_info = hw.cpuInfo
        memory_info_mb = round(hw.memorySize / (1024**2), 2)

        # Get available memory and CPU
        host_summary = host.summary
        hw_summary = host_summary.hardware
        quick_stats = host_summary.quickStats
        if quick_stats.overallCpuUsage is None:
            # This mean the host is disconnected
            continue

        # Calculate available CPU and Memory based on host.summary
        cpu_mhz_info = hw_summary.cpuMhz
        total_cpu_mhz = cpu_mhz_info * cpu_info.numCpuCores
        available_cpu_mhz = total_cpu_mhz - quick_stats.overallCpuUsage
        available_memory_mb = round(
            memory_info_mb - quick_stats.overallMemoryUsage, 2)

        vms_list = host.vm
        used_device_ids = []

        # TODO: Test support for multiple GPUs on single host
        for vm in vms_list:
            vm_status = vm.runtime.powerState
            if vm_status == 'poweredOn':
                for device in vm.config.hardware.device:
                    if isinstance(
                            device,
                            vsphere_adaptor.get_vim().vm.device.
                            VirtualPCIPassthrough):
                        if hasattr(device, 'backing') and isinstance(
                                device.backing,
                                vsphere_adaptor.get_vim().vm.device.
                                VirtualPCIPassthrough.DynamicBackingInfo,
                        ):
                            used_device_ids.append(device.backing.assignedId)

        host_info = {
            'HostName': host.name,
            'Model': system_info.model,
            'Vendor': system_info.vendor,
            'UUID': system_info.uuid,
            'TotalCPUcores': cpu_info.numCpuCores,
            'TotalCPUMHz': total_cpu_mhz,
            'CPUMHz': cpu_mhz_info,
            'TotalCPUpackages': cpu_info.numCpuPackages,
            'TotalCPUthreads': cpu_info.numCpuThreads,
            'AvailableCPUCapacityMHz': available_cpu_mhz,
            'TotalMemoryMB': memory_info_mb,
            'AvailableMemoryMB': available_memory_mb,
            'Cluster': cluster_name,
            'Datacenter': datacenter_name,
            'vCenter': vcenter_name,
            'MoRefID': host._moId,  # pylint: disable=protected-access
            'Accelerators': [],
            'GpuInfo': None,
        }
        # Retrieve PCI devices information
        pci_devices = host.hardware.pciDevice
        pci_passthru_devices = host.config.pciPassthruInfo
        for device, passthru_info in zip(pci_devices, pci_passthru_devices):
            if (decimal_to_four_char_hex(device.classId)[:2]
                    in DISPLAY_CONTROLLER_CLASS_ID_PREFIXES and
                    device.id not in VMWARE_VIRTUAL_DISPLAY_CONTROLLER_IDS):
                # Get the device status
                if passthru_info.passthruActive:
                    if device.id in used_device_ids:
                        device_status = 'Busy'
                    else:
                        device_status = 'Available'
                else:
                    device_status = 'Unavailable'
                memory_info = {'SizeInMiB': 0}
                for preset_accelerator in gpu_list:
                    if preset_accelerator['Model'].lower(
                    ) in device.deviceName.lower():
                        memory_info = {
                            'SizeInMiB': preset_accelerator['MemoryMB']
                        }
                        break
                device_info = {
                    'ID': device.id,
                    'ClassID': decimal_to_four_char_hex(device.classId),
                    'VendorID': decimal_to_four_char_hex(device.vendorId),
                    'VendorName': device.vendorName,
                    'DeviceID': decimal_to_four_char_hex(device.deviceId),
                    'DeviceName': device.deviceName,
                    'SubVendorID': decimal_to_four_char_hex(device.subVendorId),
                    'SubDeviceID': decimal_to_four_char_hex(device.subDeviceId),
                    'Status': device_status,
                    'MemorySizeMB': memory_info['SizeInMiB'],
                }
                host_info['Accelerators'].append(device_info)

                # GpuInfo is for vms.csv, which only needs one entry for each
                # host
                # Always put the GPU info into the vms.csv (
                # there are some inconsistent
                # So we should check the host's accelerators filed for the GPU
                # status
                if not host_info['GpuInfo']:
                    host_info['GpuInfo'] = {
                        'Gpus': [{
                            'Name': device.deviceName,
                            # TODO: Make count field dynamic according
                            #  to the state of the device
                            'Count': 1,
                            'Manufacturer': device.vendorName,
                            'MemoryInfo': memory_info,
                        }],
                        'TotalGpuMemoryInMiB': memory_info['SizeInMiB'],
                    }
        all_hosts_with_devices.append(host_info)
    return all_hosts_with_devices


def get_obj_by_mo_id(content, vimtype, moid):
    """    Get the vsphere managed object by moid value
    """
    obj = None
    container = content.viewManager.CreateContainerView(content.rootFolder,
                                                        vimtype, True)
    for c in container.view:
        if c._GetMoId() == moid:  # pylint: disable=protected-access
            obj = c
            break
    container.Destroy()
    return obj


def delete_object(content, mo):
    """    Deletes a vsphere managed object and
    waits for the deletion to complete
    """
    logger.info('Deleting {0}'.format(mo._GetMoId()))  # pylint: disable=protected-access
    try:
        wait_for_tasks(content, [mo.Destroy()])
        logger.info('Deleted {0}'.format(mo._GetMoId()))  # pylint: disable=protected-access
    except Exception:  # pylint: disable=broad-except
        logger.info('Unexpected error while deleting managed object {0}'.format(
            mo._GetMoId()))  # pylint: disable=protected-access
        return False
    return True


def poweron_vm(content, mo):
    """    Powers on a VM and wait for power on operation to complete
    """
    if not isinstance(mo, vsphere_adaptor.get_vim().VirtualMachine):
        return False

    logger.info('Powering on vm {0}'.format(mo._GetMoId()))  # pylint: disable=protected-access
    try:
        wait_for_tasks(content, [mo.PowerOn()])
        logger.info('{0} powered on successfully'.format(mo._GetMoId()))  # pylint: disable=protected-access
    except Exception:  # pylint: disable=broad-except
        logger.info('Unexpected error while powering on vm {0}'.format(
            mo._GetMoId()))  # pylint: disable=protected-access
        return False
    return True


def wait_internal_ip_ready(instance):
    # 4-min maximum timeout
    logger.info('Wait for IP to be ready.')
    timeout = 60 * 4
    start = time.time()
    internal_ip = None
    while True:
        if instance.summary.guest is not None:
            # The ip can still be None after this assignment
            internal_ip = instance.summary.guest.ipAddress
        if internal_ip:
            # The ip will be IPv6 at first, then be replaced by IPv4,
            # we need to wait for IPv4 which is a sign of
            # a ready instance
            if '.' in internal_ip and ':' not in internal_ip:
                logger.info(f'The IP is ready: {internal_ip}.')
                break
        if time.time() - start > timeout:
            raise RuntimeError(
                f'Failed to get the IP address after {timeout} seconds.')
        logger.debug('Retry internal ip in 5 second...')
        time.sleep(5)


def poweroff_vm(content, mo):
    """    Powers on a VM and wait for power on operation to complete
    """
    if not isinstance(mo, vsphere_adaptor.get_vim().VirtualMachine):
        return False

    logger.info('Powering off vm {0}'.format(mo._GetMoId()))  # pylint: disable=protected-access
    try:
        wait_for_tasks(content, [mo.PowerOff()])
        logger.info('{0} powered off successfully'.format(mo._GetMoId()))  # pylint: disable=protected-access
    except Exception:  # pylint: disable=broad-except
        logger.info('Unexpected error while powering off vm {0}'.format(
            mo._GetMoId()))  # pylint: disable=protected-access
        return False
    return True


def wait_for_tasks(content, tasks):
    """    Given the tasks, it returns after all the tasks are complete
    """
    task_list = [str(task) for task in tasks]

    # Create filter
    obj_specs = [
        vsphere_adaptor.get_vmodl().query.PropertyCollector.ObjectSpec(obj=task)
        for task in tasks
    ]
    prop_spec = vsphere_adaptor.get_vmodl(
    ).query.PropertyCollector.PropertySpec(type=vsphere_adaptor.get_vim().Task,
                                           pathSet=[],
                                           all=True)
    filter_spec = vsphere_adaptor.get_vmodl(
    ).query.PropertyCollector.FilterSpec()
    filter_spec.objectSet = obj_specs
    filter_spec.propSet = [prop_spec]
    task_filter = content.propertyCollector.CreateFilter(filter_spec, True)

    try:
        version, state = None, None

        # Loop looking for updates till the state moves to a completed state.
        while len(task_list):
            update = content.propertyCollector.WaitForUpdates(
                version)  # type: ignore
            for filter_set in update.filterSet:
                for obj_set in filter_set.objectSet:
                    task = obj_set.obj
                    for change in obj_set.changeSet:
                        if change.name == 'info':
                            state = change.val.state  # type: ignore
                        elif change.name == 'info.state':
                            state = change.val
                        else:
                            continue

                        if not str(task) in task_list:
                            continue

                        if state == vsphere_adaptor.get_vim(
                        ).TaskInfo.State.success:
                            # Remove task from taskList
                            task_list.remove(str(task))
                        elif state == vsphere_adaptor.get_vim(
                        ).TaskInfo.State.error:
                            raise task.info.error
            # Move to next version
            version = update.version
    finally:
        if task_filter:
            task_filter.Destroy()


def get_cluster_name_by_id(self, content, name):
    cluster_obj = get_objs_by_names(
        content, [vsphere_adaptor.get_vim().ClusterComputeResource], [name])
    if cluster_obj is not None:
        self.mo_id = cluster_obj['name']._GetMoId()  # pylint: disable=protected-access
        logger.info('Cluster MoId: {0}'.format(self.mo_id))
    else:
        logger.info('Cluster: {0} not found'.format(self.cluster_name))


def create_spec_with_script(vm, script_string: str):
    hostname = vsphere_adaptor.get_vim().vm.customization.FixedName()
    hostname.name = vm.name
    identity = vsphere_adaptor.get_vim().vm.customization.LinuxPrep(
        hostName=hostname, domain='VMWare.com', scriptText=script_string)
    global_ip_settings = vsphere_adaptor.get_vim(
    ).vm.customization.GlobalIPSettings()
    ip = vsphere_adaptor.get_vim().vm.customization.DhcpIpGenerator()
    ip_settings = vsphere_adaptor.get_vim().vm.customization.IPSettings(ip=ip)
    nic_setting_maps = []
    for device in vm.config.hardware.device:
        if isinstance(device,
                      vsphere_adaptor.get_vim().vm.device.VirtualEthernetCard):
            nic_setting_map = vsphere_adaptor.get_vim(
            ).vm.customization.AdapterMapping(adapter=ip_settings)
            nic_setting_maps.append(nic_setting_map)
    return vsphere_adaptor.get_vim().vm.customization.Specification(
        globalIPSettings=global_ip_settings,
        identity=identity,
        nicSettingMap=nic_setting_maps,
    )


def get_vm_uuid_from_vm():
    result = subprocess.run(['sudo', 'dmidecode', '-t', '1'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            text=True,
                            check=True)

    if result.returncode != 0:
        print('Error running dmidecode: ', result.stderr)
        return None

    matches = re.search(r'Serial Number: VMware-(.+)', result.stdout)

    if matches:
        serial_number = matches.group(1).strip().replace(' ',
                                                         '').replace('-', '')
        if len(serial_number) != 32:
            logger.error(f'Invalid serial number: {serial_number}')
            raise RuntimeError('Failed to get UUID from dmidecode')
        uuid = (f'{serial_number[:8]}-{serial_number[8:12]}-'
                f'{serial_number[12:16]}-{serial_number[16:20]}-'
                f'{serial_number[20:]}')
        return uuid
    else:
        logger.error('Failed to get UUID from dmidecode')
        raise RuntimeError('Failed to get UUID from dmidecode')



# pylint: disable=assignment-from-no-return
#
# The pylint exception above is an accomodation for
# false positives generated by pylint for the Vast
# python sdk.
#
"""Vast library wrapper for SkyPilot."""
from typing import Any, Dict, List, Optional

from sky import sky_logging
from sky.adaptors import vast

logger = sky_logging.init_logger(__name__)


def list_instances() -> Dict[str, Dict[str, Any]]:
    """Lists instances associated with API key."""
    instances = vast.vast().show_instances()

    instance_dict: Dict[str, Dict[str, Any]] = {}
    for instance in instances:
        instance['id'] = str(instance['id'])
        info = instance

        if isinstance(instance['actual_status'], str):
            info['status'] = instance['actual_status'].upper()
        else:
            info['status'] = 'UNKNOWN'
        info['name'] = instance['label']

        instance_dict[instance['id']] = info

    return instance_dict


def launch(name: str, instance_type: str, region: str, disk_size: int,
           image_name: str, ports: Optional[List[int]],
           preemptible: bool) -> str:
    """Launches an instance with the given parameters.

    Converts the instance_type to the Vast GPU name, finds the specs for the
    GPU, and launches the instance.

    Notes:

      *  `georegion`: This is a feature flag to provide an additional
         scope of geographical specificy while maintaining backward
         compatibility.

      *  `chunked`: This is a feature flag to give breadth to the
         snowflake nature of the vast catalog marketplace. It rounds
         down various specifications of machines to emulate an instance
         type and make them more interchangeable.

      *  `disk_size`: We look for instances that are of the requested
         size or greater than it. For instance, `disk_size=100` might
         return something with `disk_size` at 102 or even 1000.

         The disk size {xx} GB is not exactly matched the requested
         size {yy} GB. It is possible to charge extra cost on disk.

      *  `ports`: This is a feature flag to expose ports to the internet.

      *  `geolocation`: Geolocation on Vast can be as specific as the
         host chooses to be. They can say, for instance, "Yutakachō,
         Shinagawa District, Tokyo, JP." Such a specific geolocation
         as ours would fail to return this host in a simple string
         comparison if a user searched for "JP".

         Since regardless of specificity, all our geolocations end
         in two-letter country codes we just snip that to conform
         to how many providers state their geolocation.

      *  Since the catalog is cached, we can't gaurantee availability
         of any machine at the point of inquiry.  As a consequence we
         search for the machine again and potentially return a failure
         if there is no availability.

	  *  We pass in the cpu_ram here as a guarantor to make sure the
		 instance we match with will be compliant with the requested
		 amount of memory.

      *  Vast instance types are an invention for skypilot. Refer to
         catalog/vast_catalog.py for the current construction
         of the type."""
    cpu_ram = float(instance_type.split('-')[-1]) / 1024
    gpu_name = instance_type.split('-')[1].replace('_', ' ')
    num_gpus = int(instance_type.split('-')[0].replace('x', ''))

    query = ' '.join([
        'chunked=true',
        'georegion=true',
        f'geolocation="{region[-2:]}"',
        f'disk_space>={disk_size}',
        f'num_gpus={num_gpus}',
        f'gpu_name="{gpu_name}"',
        f'cpu_ram>="{cpu_ram}"',
    ])

    instance_list = vast.vast().search_offers(query=query)

    if isinstance(instance_list, int) or len(instance_list) == 0:
        raise RuntimeError('Failed to create instances, could not find an '
                           f'offer that satisfies the requirements "{query}".')

    instance_touse = instance_list[0]

    port_map = ' '.join([f'-p {p}:{p}' for p in ports]) if ports else ''

    launch_params = {
        'id': instance_touse['id'],
        'direct': True,
        'ssh': True,
        'env': f'-e __SOURCE=skypilot {port_map}',
        'onstart_cmd': ';'.join([
            'touch ~/.no_auto_tmux',
            f'echo "{vast.vast().api_key_access}" > ~/.vast_api_key',
        ]),
        'label': name,
        'image': image_name,
        'disk': disk_size
    }

    if preemptible:
        launch_params['min_bid'] = instance_touse['min_bid']

    new_instance_contract = vast.vast().create_instance(**launch_params)

    new_instance = vast.vast().show_instance(
        id=new_instance_contract['new_contract'])

    return new_instance['id']


def start(instance_id: str) -> None:
    """Starts the given instance."""
    vast.vast().start_instance(id=instance_id)


def stop(instance_id: str) -> None:
    """Stops the given instance."""
    vast.vast().stop_instance(id=instance_id)


def remove(instance_id: str) -> None:
    """Terminates the given instance."""
    vast.vast().destroy_instance(id=instance_id)


def get_ssh_ports(cluster_name: str) -> List[int]:
    """Gets the SSH ports for the given cluster."""
    logger.debug(f'Getting SSH ports for cluster {cluster_name}.')

    instances = list_instances()
    possible_names = [f'{cluster_name}-head', f'{cluster_name}-worker']

    ssh_ports = []

    for instance in instances.values():
        if instance['name'] in possible_names:
            ssh_ports.append(instance['ssh_port'])
    assert ssh_ports, (
        f'Could not find any instances for cluster {cluster_name}.')

    return ssh_ports
